@@ -12,6 +12,7 @@ from functools import reduce
 # pylint: disable=no-name-in-module
 from mininet.log import error, debug
 from mininet.node import Host
+from mininet.util import quietRun, errRun
 
 from mininet_test_util import DEVNULL
 
@@ -63,8 +64,19 @@ class DockerHost(Host):
         self.vol_maps = vol_maps
         self.network = network
         self.startup_timeout_ms = startup_timeout_ms
+        self.name = name
+        self.pullImage()
         Host.__init__(self, name, **kwargs)
 
+    def pullImage(self): # pylint: disable=invalid-name
+        "Pull docker image if necessary"
+        if self.image not in quietRun('docker images'):
+            error('%s: docker image' % self.name, self.image,
+                  'not available locally - pulling\n')
+            _out, err, code = errRun('docker', 'pull', self.image)
+            if err or code:
+                error('docker pull failed with error', code, err, '\n')
+
     # pylint: disable=invalid-name
     def startShell(self, mnopts=None):
         """Start a shell process for running commands."""
@@ -96,7 +108,7 @@ class DockerHost(Host):
         debug('docker command "%s", fd %d, fd %d' % (' '.join(cmd), self.master, self.slave))
         try:
             self.shell = self._popen(cmd, stdin=self.slave, stdout=self.slave, stderr=self.slave)
-            self.stdin = os.fdopen(self.master, 'rw')
+            self.stdin = os.fdopen(self.master, 'r')
             self.stdout = self.stdin
             self.pollOut = select.poll() # pylint: disable=invalid-name
             self.pollOut.register(self.stdout) # pylint: disable=no-member
@@ -171,7 +183,7 @@ class DockerHost(Host):
             config_json = inspect_pipe.stdout.readlines()
             inspect_pipe.stdout.close()
             assert len(config_json) == 1, "Expected 1 config line, found %s" % len(config_json)
-            config = json.loads(config_json[0])
+            config = json.loads(config_json[0].decode())
             entryconfig = config['Entrypoint']
             entrypoint = entryconfig if entryconfig else ['/usr/bin/env']
             cmd = config['Cmd'] if 'Cmd' in config else []
