@@ -9,10 +9,10 @@ import shutil
 import subprocess
 import time
 
-import netifaces
-
 # pylint: disable=import-error
 # pylint: disable=no-name-in-module
+# pylint: disable=too-many-arguments
+
 from mininet.log import error, output
 from mininet.topo import Topo
 from mininet.node import Controller
@@ -21,6 +21,8 @@ from mininet.node import OVSSwitch
 
 import mininet_test_util
 
+import netifaces
+
 # TODO: mininet 2.2.2 leaks ptys (master slave assigned in startShell)
 # override as necessary close them. Transclude overridden methods
 # to avoid multiple inheritance complexity.
@@ -58,7 +60,7 @@ class FaucetHostCleanup(object):
         self.shell = self._popen( # pylint: disable=no-member; # pytype: disable=attribute-error
             cmd, stdin=self.slave, stdout=self.slave, stderr=self.slave,
             close_fds=False)
-        self.stdin = os.fdopen(self.master, 'rw')
+        self.stdin = os.fdopen(self.master, 'r')
         self.stdout = self.stdin
         self.pid = self.shell.pid
         self.pollOut = select.poll() # pylint: disable=invalid-name
@@ -151,6 +153,13 @@ class FaucetSwitch(FaucetHostCleanup, OVSSwitch):
                 self.TCReapply(intf)
 
 
+class NoControllerFaucetSwitch(FaucetSwitch):
+    """A switch without any controllers (typically for remapping hardware to software."""
+
+    def start(self, _controllers):
+        super(NoControllerFaucetSwitch, self).start(controllers=[])
+
+
 class VLANHost(FaucetHost):
     """Implementation of a Mininet host on a tagged VLAN."""
 
@@ -186,7 +195,7 @@ class FaucetSwitchTopo(Topo):
         """Return a unique switch/host prefix for a test."""
         # Linux tools require short interface names.
         # pylint: disable=no-member
-        id_chars = string.letters + string.digits # pytype: disable=module-attr
+        id_chars = string.ascii_letters + string.digits # pytype: disable=module-attr
         id_a = int(ports_served / len(id_chars))
         id_b = ports_served - (id_a * len(id_chars))
         return '%s%s' % (
@@ -208,12 +217,12 @@ class FaucetSwitchTopo(Topo):
         host_name = 'e%s%1.1u' % (sid_prefix, host_n + 1)
         return self.addHost(name=host_name, cls=e_cls, host_n=host_n, tmpdir=tmpdir)
 
-    def _add_faucet_switch(self, sid_prefix, dpid, ovs_type):
+    def _add_faucet_switch(self, sid_prefix, dpid, ovs_type, switch_cls):
         """Add a FAUCET switch."""
         switch_name = 's%s' % sid_prefix
         return self.addSwitch(
             name=switch_name,
-            cls=FaucetSwitch,
+            cls=switch_cls,
             datapath=ovs_type,
             dpid=mininet_test_util.mininet_dpid(dpid))
 
@@ -236,12 +245,14 @@ class FaucetSwitchTopo(Topo):
                 self._add_untagged_host(sid_prefix, host_n)
             for host_n in range(n_extended):
                 self._add_extended_host(sid_prefix, host_n, e_cls, tmpdir)
+            switch_cls = FaucetSwitch
             if hw_dpid and hw_dpid == dpid:
                 remap_dpid = str(int(dpid) + 1)
                 output('bridging hardware switch DPID %s (%x) dataplane via OVS DPID %s (%x)' % (
                     dpid, int(dpid), remap_dpid, int(remap_dpid)))
                 dpid = remap_dpid
-            switch = self._add_faucet_switch(sid_prefix, dpid, ovs_type)
+                switch_cls = NoControllerFaucetSwitch
+            switch = self._add_faucet_switch(sid_prefix, dpid, ovs_type, switch_cls)
             self._add_links(switch, self.hosts(), links_per_host)
 
 
@@ -253,7 +264,7 @@ class FaucetStringOfDPSwitchTopo(FaucetSwitchTopo):
     def build(self, ovs_type, ports_sock, test_name, dpids,
               n_tagged=0, tagged_vid=100, n_untagged=0,
               links_per_host=0, switch_to_switch_links=1,
-              hw_dpid=None):
+              hw_dpid=None, stack_ring=False):
         """
 
                                Hosts
@@ -279,6 +290,11 @@ class FaucetStringOfDPSwitchTopo(FaucetSwitchTopo):
         * (n_tagged + n_untagged + 2) links on switches 0 < n < s-1,
           with final two links being inter-switch
         """
+        def addLinks(src, dst):
+            for _ in range(self.switch_to_switch_links):
+                self.addLink(src, dst)
+
+        first_switch = None
         last_switch = None
         self.switch_to_switch_links = switch_to_switch_links
         for dpid in dpids:
@@ -290,19 +306,24 @@ class FaucetStringOfDPSwitchTopo(FaucetSwitchTopo):
                 hosts.append(self._add_tagged_host(sid_prefix, tagged_vid, host_n))
             for host_n in range(n_untagged):
                 hosts.append(self._add_untagged_host(sid_prefix, host_n))
+            switch_cls = FaucetSwitch
             if hw_dpid and hw_dpid == dpid:
                 remap_dpid = str(int(dpid) + 1)
                 output('bridging hardware switch DPID %s (%x) dataplane via OVS DPID %s (%x)' % (
                     dpid, int(dpid), remap_dpid, int(remap_dpid)))
                 dpid = remap_dpid
-            switch = self._add_faucet_switch(sid_prefix, dpid, ovs_type)
+                switch_cls = NoControllerFaucetSwitch
+            switch = self._add_faucet_switch(sid_prefix, dpid, ovs_type, switch_cls)
             self._add_links(switch, hosts, links_per_host)
+            if first_switch is None:
+                first_switch = switch
             if last_switch is not None:
                 # Add a switch-to-switch link with the previous switch,
                 # if this isn't the first switch in the topology.
-                for _ in range(self.switch_to_switch_links):
-                    self.addLink(last_switch, switch)
+                addLinks(last_switch, switch)
             last_switch = switch
+        if stack_ring:
+            addLinks(first_switch, last_switch)
 
 
 class BaseFAUCET(Controller):
@@ -518,7 +539,7 @@ class FAUCET(BaseFAUCET):
         self.ofctl_port = mininet_test_util.find_free_port(
             ports_sock, test_name)
         cargs = ' '.join((
-            '--ryu-wsapi-host=%s' % mininet_test_util.LOCALHOST,
+            '--ryu-wsapi-host=%s' % mininet_test_util.LOCALHOSTV6,
             '--ryu-wsapi-port=%u' % self.ofctl_port,
             self._tls_cargs(port, ctl_privkey, ctl_cert, ca_certs)))
         super(FAUCET, self).__init__(
@@ -556,4 +577,4 @@ class Gauge(BaseFAUCET):
 class FaucetExperimentalAPI(FAUCET):
     """Start a controller to run the Faucet experimental API tests."""
 
-    START_ARGS = ['--ryu-app=test_experimental_api.py', '--ryu-app=ryu.app.ofctl_rest']
+    START_ARGS = ['--ryu-app=experimental_api_test_app.py', '--ryu-app=ryu.app.ofctl_rest']
