@@ -4,29 +4,37 @@ UNITTESTS=1
 DEPCHECK=1
 MINCOVERAGE=85
 
-cd /faucet-src
-
-if [ -d /var/tmp/pip-cache ] ; then
-  cp -r /var/tmp/pip-cache /var/tmp/pip-cache-local || exit 1
-fi
-./docker/pip_deps.sh "--cache-dir=/var/tmp/pip-cache-local" || exit 1
-
-# if -n passed, don't check dependencies/lint/type/documentation.
-# wrapper script only cares about -n, others passed to test suite.
-while getopts "cdjknsxi" o $FAUCET_TESTS; do
+# allow user to skip parts of docker test
+# this wrapper script only cares about -n, -u, -i, others passed to test suite.
+while getopts "cdijknsux" o $FAUCET_TESTS; do
   case "${o}" in
-        n)
-            DEPCHECK=0
-            ;;
         i)
+            # run only integration tests
             UNITTESTS=0
             DEPCHECK=0
             ;;
+        n)
+            # skip code check
+            DEPCHECK=0
+            ;;
+        u)
+            # skip unit tests
+            UNITTESTS=0
+            ;;
         *)
             ;;
     esac
 done
 
+cd /faucet-src
+
+if [ -d /var/tmp/pip-cache ] ; then
+  echo Using pip cache
+  cp -r /var/tmp/pip-cache /var/tmp/pip-cache-local || exit 1
+fi
+./docker/pip_deps.sh "--cache-dir=/var/tmp/pip-cache-local" || exit 1
+./docker/workarounds.sh || exit 1
+
 echo "========== checking IPv4/v6 localhost is up ====="
 ping6 -c 1 ::1 || exit 1
 ping -c 1 127.0.0.1 || exit 1
@@ -43,12 +51,13 @@ cd /faucet-src/tests
 
 ./sysctls_for_tests.sh
 
+# TODO: need to force UTF-8 as POSIX causes python3/pytype errors.
+locale-gen en_US.UTF-8 || exit 1
+
 if [ "$UNITTESTS" == 1 ] ; then
     echo "========== Running faucet unit tests =========="
     cd /faucet-src/tests
-    PYTHONPATH=.. ./test_coverage.sh || exit 1
-    # TODO: enable under travis
-    # codecov || true
+    LANG=en_US.UTF-8 LANGUAGE=en_US.en LC_ALL=en_US.UTF-8 ./run_unit_tests.sh || exit 1
 fi
 
 if [ "$DEPCHECK" == 1 ] ; then
@@ -57,14 +66,12 @@ if [ "$DEPCHECK" == 1 ] ; then
     make html || exit 1
     rm -rf _build
 
-    cd /faucet-src/tests
+    cd /faucet-src/tests/codecheck
     echo "============ Running pytype analyzer ============"
-    # TODO: need to force UTF-8 as POSIX causes pytype errors
-    locale-gen en_US.UTF-8 || exit 1
-    LANG=en_US.UTF-8 LANGUAGE=en_US.en LC_ALL=en_US.UTF-8 ./test_pytype.sh || exit 1
+    LANG=en_US.UTF-8 LANGUAGE=en_US.en LC_ALL=en_US.UTF-8 ./pytype.sh || exit 1
 
     echo "============ Running pylint analyzer ============"
-    PYTHONPATH=.. ./test_min_pylint.sh || exit 1
+    PYTHONPATH=../.. ./pylint.sh || exit 1
 fi
 
 echo "========== Starting docker container =========="
@@ -72,14 +79,15 @@ service docker start
 
 echo "========== Running faucet system tests =========="
 test_failures=
-export PYTHONPATH=/faucet-src
+export FAUCET_DIR=/faucet-src/faucet
+export PYTHONPATH=/faucet-src:/faucet-src/faucet:/faucet-src/clib
 
-cd /faucet-src/tests
-python2 ./faucet_mininet_test.py -c
-http_proxy="" python2 ./faucet_mininet_test.py $FAUCET_TESTS || test_failures+=" faucet_mininet_test"
+cd /faucet-src/tests/integration
+./mininet_main.py -c
+LANG=en_US.UTF-8 LANGUAGE=en_US.en LC_ALL=en_US.UTF-8 http_proxy="" ./mininet_main.py $FAUCET_TESTS || test_failures+=" mininet_main"
 
 cd /faucet-src/clib
-http_proxy="" python2 ./clib_mininet_test.py $FAUCET_TESTS || test_failures+=" clib_mininet_test"
+LANG=en_US.UTF-8 LANGUAGE=en_US.en LC_ALL=en_US.UTF-8 http_proxy="" ./clib_mininet_test.py $FAUCET_TESTS || test_failures+=" clib_mininet_test"
 
 if [ -n "$test_failures" ]; then
     echo Test failures: $test_failures
