@@ -16,24 +16,24 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from collections import defaultdict
 import copy
+from collections import defaultdict
+import random
 import netaddr
 
 from datadiff import diff
-from netaddr.core import AddrFormatError
 import networkx
 
 from faucet import faucet_pipeline
-from faucet import valve_acl
 from faucet import valve_of
-from faucet.conf import Conf, InvalidConfigError, test_config_condition
+from faucet.conf import Conf, test_config_condition
+from faucet.valve import SUPPORTED_HARDWARE
+from faucet.faucet_pipeline import ValveTableConfig
 from faucet.valve_table import ValveTable, ValveGroupTable
-from faucet.valve_util import get_setting
-from faucet.valve_packet import FAUCET_MAC
+from faucet import valve_packet
 
 
-class NullRyuDatapath(object):
+class NullRyuDatapath:
     """Placeholder Ryu Datapath."""
 
     ofproto = valve_of.ofp
@@ -48,54 +48,7 @@ class DP(Conf):
 configuration.
 """
 
-    acls = None
-    vlans = None
-    interfaces = None # config
-    interface_ranges = None
-    ports = None
-    routers = None
-    running = False
-    name = None
-    dp_id = None
-    cookie = None
-    configured = False
-    priority_offset = None
-    low_priority = None
-    high_priority = None
-    stack = None
-    stack_ports = None
-    output_only_ports = None
-    ignore_learn_ins = None
-    drop_broadcast_source_address = None
-    drop_spoofed_faucet_mac = None
-    groups = None
-    group_table = False
-    group_table_routing = False
-    max_hosts_per_resolve_cycle = None
-    max_host_fib_retry_count = None
-    max_resolve_backoff_time = None
-    packetin_pps = None
-    learn_jitter = None
-    learn_ban_timeout = None
-    advertise_interval = None
-    proactive_learn = None
-    pipeline_config_dir = None
-    use_idle_timeout = None
-    tables = {} # type: dict
-    tables_by_id = {} # type: dict
-    meters = {} # type: dict
-    timeout = None
-    arp_neighbor_timeout = None
-    lldp_beacon = {} # type: dict
-    metrics_rate_limit_sec = None
-    faucet_dp_mac = None
-    combinatorial_port_flood = None
-    lacp_timeout = None
-    dp_acls = None
-    dot1x = None
-
-    dyn_last_coldstart_time = None
-    dyn_up_ports = set() # type: ignore
+    mutable_attrs = frozenset(['stack', 'vlans'])
 
     # Values that are set to None will be set using set_defaults
     # they are included here for testing and informational purposes
@@ -138,8 +91,6 @@ configuration.
         # By default drop packets on datapath spoofing the FAUCET_MAC
         'group_table': False,
         # Use GROUP tables for VLAN flooding
-        'group_table_routing': False,
-        # Use GROUP tables for routing (nexthops)
         'max_hosts_per_resolve_cycle': 5,
         # Max hosts to try to resolve per gateway resolution cycle.
         'max_host_fib_retry_count': 10,
@@ -154,17 +105,17 @@ configuration.
         # When banning/limiting learning, wait this many seconds before learning can be retried
         'advertise_interval': 30,
         # How often to advertise (eg. IPv6 RAs)
-        'proactive_learn': True,
-        # whether proactive learning is enabled for IP nexthops
-        'pipeline_config_dir': get_setting('FAUCET_PIPELINE_DIR', True),
-        # where config files for pipeline are stored (if any).
+        'proactive_learn_v4': True,
+        # whether proactive learning is enabled for IPv4 nexthops
+        'proactive_learn_v6': True,
+        # whether proactive learning is enabled for IPv6 nexthops
         'use_idle_timeout': False,
         # Turn on/off the use of idle timeout for src_table, default OFF.
         'lldp_beacon': {},
         # Config for LLDP beacon service.
         'metrics_rate_limit_sec': 0,
         # Rate limit metric updates - don't update metrics if last update was less than this many seconds ago.
-        'faucet_dp_mac': FAUCET_MAC,
+        'faucet_dp_mac': valve_packet.FAUCET_MAC,
         # MAC address of packets sent by FAUCET, not associated with any VLAN.
         'combinatorial_port_flood': False,
         # if True, use a seperate output flow for each input port on this VLAN.
@@ -174,6 +125,14 @@ configuration.
         # List of dataplane ACLs (overriding per port ACLs).
         'dot1x': None,
         # Experimental dot1x configuration.
+        'table_sizes': {},
+        # Table sizes for TFM switches.
+        'min_wildcard_table_size': 32,
+        # Minimum table size for wildcard tables.
+        'max_wildcard_table_size': 1024 + 256,
+        # Maximum table size for wildcard tables.
+        'global_vlan': 0,
+        # Reserved VID for internal global router VLAN.
         }
 
     defaults_types = {
@@ -199,7 +158,6 @@ configuration.
         'drop_bpdu': bool,
         'drop_lldp': bool,
         'group_table': bool,
-        'group_table_routing': bool,
         'max_hosts_per_resolve_cycle': int,
         'max_host_fib_retry_count': int,
         'max_resolve_backoff_time': int,
@@ -207,8 +165,8 @@ configuration.
         'learn_jitter': int,
         'learn_ban_timeout': int,
         'advertise_interval': int,
-        'proactive_learn': bool,
-        'pipeline_config_dir': str,
+        'proactive_learn_v4': bool,
+        'proactive_learn_v6': bool,
         'use_idle_timeout': bool,
         'lldp_beacon': dict,
         'metrics_rate_limit_sec': int,
@@ -216,6 +174,22 @@ configuration.
         'combinatorial_port_flood': bool,
         'dp_acls': list,
         'dot1x': dict,
+        'table_sizes': dict,
+        'min_wildcard_table_size': int,
+        'max_wildcard_table_size': int,
+        'global_vlan': int,
+    }
+
+    default_table_sizes_types = {
+        'port_acl': int,
+        'vlan': int,
+        'vlan_acl': int,
+        'eth_src': int,
+        'ipv4_fib': int,
+        'ipv6_fib': int,
+        'vip': int,
+        'eth_dst': int,
+        'flood': int,
     }
 
     stack_defaults_types = {
@@ -233,12 +207,71 @@ configuration.
     }
 
     wildcard_table = ValveTable(
-        valve_of.ofp.OFPTT_ALL, 'all', None, flow_cookie=0)
+        valve_of.ofp.OFPTT_ALL, 'all', ValveTableConfig('all'), flow_cookie=0)
 
 
     def __init__(self, _id, dp_id, conf):
         """Constructs a new DP object"""
-        super(DP, self).__init__(_id, dp_id, conf)
+        self.acls = None
+        self.acls_in = None
+        self.advertise_interval = None
+        self.arp_neighbor_timeout = None
+        self.bgp_local_address = None
+        self.bgp_neighbor_as = None
+        self.bgp_routerid = None
+        self.combinatorial_port_flood = None
+        self.configured = False
+        self.cookie = None
+        self.description = None
+        self.dot1x = None
+        self.dp_acls = None
+        self.dp_id = None
+        self.drop_broadcast_source_address = None
+        self.drop_spoofed_faucet_mac = None
+        self.dyn_last_coldstart_time = None
+        self.dyn_running = False
+        self.dyn_up_ports = None
+        self.faucet_dp_mac = None
+        self.global_vlan = None
+        self.groups = None
+        self.group_table = False
+        self.hardware = None
+        self.high_priority = None
+        self.highest_priority = None
+        self.ignore_learn_ins = None
+        self.interface_ranges = None
+        self.interfaces = None
+        self.lacp_timeout = None
+        self.learn_ban_timeout = None
+        self.learn_jitter = None
+        self.lldp_beacon = None
+        self.low_priority = None
+        self.lowest_priority = None
+        self.max_host_fib_retry_count = None
+        self.max_hosts_per_resolve_cycle = None
+        self.max_resolve_backoff_time = None
+        self.meters = None
+        self.metrics_rate_limit_sec = None
+        self.name = None
+        self.ofchannel_log = None
+        self.output_only_ports = None
+        self.packetin_pps = None
+        self.ports = None
+        self.priority_offset = None
+        self.proactive_learn_v4 = None
+        self.proactive_learn_v6 = None
+        self.proactive_nd_limit = None
+        self.routers = None
+        self.stack = None
+        self.stack_ports = None
+        self.tables = None
+        self.timeout = None
+        self.unicast_flood = None
+        self.use_idle_timeout = None
+        self.vlans = None
+        self.min_wildcard_table_size = None
+        self.max_wildcard_table_size = None
+
         self.acls = {}
         self.vlans = {}
         self.ports = {}
@@ -246,19 +279,24 @@ configuration.
         self.stack_ports = []
         self.output_only_ports = []
         self.lldp_beacon_ports = []
+        self.tables = {}
+        self.meters = {}
+        self.lldp_beacon = {}
+        self.table_sizes = {}
+        self.dyn_up_ports = set()
+        super(DP, self).__init__(_id, dp_id, conf)
 
     def __str__(self):
         return self.name
 
     def check_config(self):
+        super(DP, self).check_config()
         test_config_condition(not isinstance(self.dp_id, int), (
             'dp_id must be %s not %s' % (int, type(self.dp_id))))
         test_config_condition(self.dp_id < 0 or self.dp_id > 2**64-1, (
             'DP ID %s not in valid range' % self.dp_id))
         test_config_condition(not netaddr.valid_mac(self.faucet_dp_mac), (
             'invalid MAC address %s' % self.faucet_dp_mac))
-        test_config_condition(self.group_table and self.group_table_routing, (
-            'groups for routing and other functions simultaneously not supported'))
         test_config_condition(not (self.interfaces or self.interface_ranges), (
             'DP %s must have at least one interface' % self))
         # To prevent L2 learning from timing out before L3 can refresh
@@ -278,16 +316,32 @@ configuration.
             self._check_conf_types(self.stack, self.stack_defaults_types)
         if self.dot1x:
             self._check_conf_types(self.dot1x, self.dot1x_defaults_types)
+        self._check_conf_types(self.table_sizes, self.default_table_sizes_types)
 
-    def _configure_tables(self):
-        """Configure FAUCET pipeline of tables with matches."""
+    def _configure_tables(self, override_table_config, valve_cl, vlan_port_factor):
+        """Configure FAUCET pipeline with tables."""
+        tables = {}
         self.groups = ValveGroupTable()
+        relative_table_id = 0
         for table_id, table_config in enumerate(faucet_pipeline.FAUCET_PIPELINE):
-            table_name, restricted_match_types = table_config
-            self.tables[table_name] = ValveTable(
-                table_id, table_name, restricted_match_types,
-                self.cookie, notify_flow_removed=self.use_idle_timeout)
-            self.tables_by_id[table_id] = self.tables[table_name]
+            table_name = table_config.name
+            if table_name in override_table_config:
+                table_config = override_table_config[table_name]
+            size = self.table_sizes.get(table_name, self.min_wildcard_table_size)
+            if table_config.vlan_port_scale:
+                size = max(size, int(vlan_port_factor * float(table_config.vlan_port_scale)))
+            if not table_config.exact_match:
+                size = min(size, self.max_wildcard_table_size)
+                size = int(size / self.min_wildcard_table_size) * self.min_wildcard_table_size
+            table_config.size = size
+            if table_config.match_types:
+                if not valve_cl.STATIC_TABLE_IDS:
+                    table_id = relative_table_id
+                tables[table_name] = ValveTable(
+                    table_id, table_name, table_config, self.cookie,
+                    notify_flow_removed=self.use_idle_timeout)
+                relative_table_id += 1
+        self.tables = tables
 
     def set_defaults(self):
         super(DP, self).set_defaults()
@@ -298,14 +352,19 @@ configuration.
         self._set_default('high_priority', self.low_priority + 1)
         self._set_default('highest_priority', self.high_priority + 98)
         self._set_default('description', self.name)
-        self._configure_tables()
+
+    def table_by_id(self, table_id):
+        tables = [table for table in list(self.tables.values()) if table_id == table.table_id]
+        if tables:
+            return tables[0]
+        return None
 
     def match_tables(self, match_type):
         """Return list of tables with matches of a specific match type."""
         match_tables = []
-        for table in list(self.tables_by_id.values()):
-            if table.restricted_match_types is not None:
-                if match_type in table.restricted_match_types:
+        for table in list(self.tables.values()):
+            if table.match_types is not None:
+                if match_type in table.match_types:
                     match_tables.append(table)
             else:
                 match_tables.append(table)
@@ -315,14 +374,6 @@ configuration.
         """Return list of tables that specify in_port as a match."""
         return self.match_tables('in_port')
 
-    def vlan_match_tables(self):
-        """Return list of tables that specify vlan_vid as a match."""
-        return self.match_tables('vlan_vid')
-
-    def all_valve_tables(self):
-        """Return list of all Valve tables."""
-        return list(self.tables_by_id.values())
-
     def add_acl(self, acl_ident, acl):
         """Add an ACL to this DP."""
         self.acls[acl_ident] = acl
@@ -342,8 +393,28 @@ configuration.
         if port.lldp_beacon_enabled():
             self.lldp_beacon_ports.append(port)
 
-    def resolve_stack_topology(self, dps):
-        """Resolve inter-DP config for stacking."""
+    def lldp_beacon_send_ports(self, now):
+        """Return list of ports to send LLDP packets; stacked ports always send LLDP."""
+        send_ports = []
+        if self.lldp_beacon:
+            priority_ports = {
+                port for port in self.stack_ports
+                if port.running() and port.lldp_beacon_enabled()}
+            cutoff_beacon_time = now - self.lldp_beacon['send_interval']
+            nonpriority_ports = {
+                port for port in self.lldp_beacon_ports
+                if port.running() and (
+                    port.dyn_last_lldp_beacon_time is None or
+                    port.dyn_last_lldp_beacon_time < cutoff_beacon_time)}
+            nonpriority_ports -= priority_ports
+            send_ports.extend(list(priority_ports))
+            send_ports.extend(list(nonpriority_ports)[:self.lldp_beacon['max_per_interval']])
+            random.shuffle(send_ports)
+        return send_ports
+
+    @staticmethod
+    def modify_stack_topology(graph, dp, port, add=True):
+        """Add/remove an edge to the stack graph which originates from this dp and port."""
 
         def canonical_edge(dp, port):
             peer_dp = port.stack['dp']
@@ -370,6 +441,33 @@ configuration.
                 'dp_a': edge_a_dp, 'port_a': edge_a_port,
                 'dp_z': edge_z_dp, 'port_z': edge_z_port}
 
+        edge = canonical_edge(dp, port)
+        edge_a, edge_z = edge
+        edge_name = make_edge_name(edge_a, edge_z)
+        edge_attr = make_edge_attr(edge_a, edge_z)
+        edge_a_dp, _ = edge_a
+        edge_z_dp, _ = edge_z
+        if add:
+            graph.add_edge(
+                edge_a_dp.name, edge_z_dp.name,
+                key=edge_name, port_map=edge_attr)
+        elif (edge_a_dp.name, edge_z_dp.name, edge_name) in graph.edges:
+            graph.remove_edge(edge_a_dp.name, edge_z_dp.name, edge_name)
+
+        return edge_name
+
+    @classmethod
+    def add_stack_link(cls, graph, dp, port):
+        """Add a stack link to the stack graph."""
+        return cls.modify_stack_topology(graph, dp, port)
+
+    @classmethod
+    def remove_stack_link(cls, graph, dp, port):
+        """Remove a stack link to the stack graph."""
+        return cls.modify_stack_topology(graph, dp, port, False)
+
+    def resolve_stack_topology(self, dps):
+        """Resolve inter-DP config for stacking."""
         root_dp = None
         stack_dps = []
         for dp in dps:
@@ -381,7 +479,7 @@ configuration.
                     test_config_condition(root_dp is not None, 'cannot have multiple stack roots')
                     root_dp = dp
                     for vlan in list(dp.vlans.values()):
-                        test_config_condition(vlan.faucet_vips != [], (
+                        test_config_condition(vlan.faucet_vips, (
                             'routing + stacking not supported'))
 
         if root_dp is None:
@@ -395,18 +493,10 @@ configuration.
             if dp.stack_ports:
                 graph.add_node(dp.name)
                 for port in dp.stack_ports:
-                    edge = canonical_edge(dp, port)
-                    edge_a, edge_z = edge
-                    edge_name = make_edge_name(edge_a, edge_z)
-                    edge_attr = make_edge_attr(edge_a, edge_z)
-                    edge_a_dp, _ = edge_a
-                    edge_z_dp, _ = edge_z
+                    edge_name = self.add_stack_link(graph, dp, port)
                     if edge_name not in edge_count:
                         edge_count[edge_name] = 0
                     edge_count[edge_name] += 1
-                    graph.add_edge(
-                        edge_a_dp.name, edge_z_dp.name,
-                        key=edge_name, port_map=edge_attr)
         if graph.size():
             for edge_name, count in list(edge_count.items()):
                 test_config_condition(count != 2, '%s defined only in one direction' % edge_name)
@@ -415,12 +505,25 @@ configuration.
                     self.stack = {}
                 self.stack['root_dp'] = root_dp
                 self.stack['graph'] = graph
+                longest_path_to_root_len = 0
+                for dp in graph.nodes():
+                    path_to_root_len = len(self.shortest_path(root_dp.name, src_dp=dp))
+                    test_config_condition(
+                        path_to_root_len == 0, '%s not connected to stack' % dp)
+                    longest_path_to_root_len = max(
+                        path_to_root_len, longest_path_to_root_len)
+                self.stack['longest_path_to_root_len'] = longest_path_to_root_len
 
-    def shortest_path(self, dest_dp):
+    def shortest_path(self, dest_dp, src_dp=None):
         """Return shortest path to a DP, as a list of DPs."""
+        if src_dp is None:
+            src_dp = self.name
         if self.stack is not None and 'root_dp' in self.stack:
-            return networkx.shortest_path(
-                self.stack['graph'], self.name, dest_dp)
+            try:
+                return networkx.shortest_path(
+                    self.stack['graph'], src_dp, dest_dp)
+            except networkx.exception.NetworkXNoPath:
+                pass
         return []
 
     def shortest_path_to_root(self):
@@ -432,6 +535,16 @@ configuration.
                 return self.shortest_path(root_dp.name)
         return []
 
+    def is_stack_root(self):
+        """Return True if this DP is the root of the stack."""
+        return 'priority' in self.stack
+
+    def is_stack_edge(self):
+        """Return True if this DP is a stack edge."""
+        if self.stack and 'longest_path_to_root_len' in self.stack:
+            return self.stack['longest_path_to_root_len'] == len(self.shortest_path_to_root())
+        return False
+
     def peer_stack_up_ports(self, peer_dp):
         """Return list of stack ports that are up towards a peer."""
         return [port for port in self.stack_ports if port.running() and (
@@ -440,7 +553,7 @@ configuration.
     def shortest_path_port(self, dest_dp):
         """Return first port on our DP, that is the shortest path towards dest DP."""
         shortest_path = self.shortest_path(dest_dp)
-        if shortest_path is not None:
+        if len(shortest_path) > 1:
             peer_dp = shortest_path[1]
             peer_dp_ports = self.peer_stack_up_ports(peer_dp)
             if peer_dp_ports:
@@ -482,17 +595,13 @@ configuration.
                     resolved_ports.append(port)
             return resolved_ports
 
-        def resolve_port_numbers(port_names):
-            """Resolve list of ports to numbers, by port by name or number."""
-            return [port.number for port in resolve_ports(port_names)]
-
         def resolve_vlan(vlan_name):
             """Resolve VLAN by name or VID."""
             test_config_condition(not isinstance(vlan_name, (str, int)), (
                 'VLAN must be type %s or %s not %s' % (str, int, type(vlan_name))))
             if vlan_name in vlan_by_name:
                 return vlan_by_name[vlan_name]
-            elif vlan_name in self.vlans:
+            if vlan_name in self.vlans:
                 return self.vlans[vlan_name]
             return None
 
@@ -539,144 +648,113 @@ configuration.
                         'override_output_port port not defined'))
                     self.ports[port_no] = port
 
-        def resolve_acl(acl_in):
+        def resolve_acl(acl_in, vid):
             """Resolve an individual ACL."""
             test_config_condition(acl_in not in self.acls, (
-                'missing ACL %s on %s' % (self.name, acl_in)))
+                'missing ACL %s in DP: %s' % (acl_in, self.name)))
             acl = self.acls[acl_in]
-            mirror_destinations = set()
+            def resolve_port_cb(port_name):
+                port = self.resolve_port(port_name)
+                if port:
+                    return port.number
+                return port
 
-            def resolve_meter(_acl, action_conf):
-                meter_name = action_conf
+            acl.resolve_ports(resolve_port_cb)
+
+            for meter_name in acl.get_meters():
                 test_config_condition(meter_name not in self.meters, (
                     'meter %s is not configured' % meter_name))
-                return action_conf
-
-            def resolve_mirror(_acl, action_conf):
-                port_name = action_conf
-                port = self.resolve_port(port_name)
-                # If this DP does not have this port, do nothing.
-                if port is not None:
-                    action_conf = port.number
-                    mirror_destinations.add(port.number)
-                    return action_conf
-                return None
-
-            def resolve_output(_acl, action_conf):
-                resolved_action_conf = {}
-                test_config_condition(not isinstance(action_conf, dict), (
-                    'action conf is not a dictionary'))
-                for output_action, output_action_values in list(action_conf.items()):
-                    if output_action == 'port':
-                        port_name = output_action_values
-                        port = self.resolve_port(port_name)
-                        test_config_condition(not port, (
-                            'output port not defined in DP: %s' % self.name))
-                        resolved_action_conf[output_action] = port.number # pytype: disable=attribute-error
-                    elif output_action == 'ports':
-                        resolved_ports = resolve_port_numbers(output_action_values)
-                        test_config_condition(len(resolved_ports) != len(output_action_values), (
-                            'output port(s) not defined in DP: %s' % self.name))
-                        resolved_action_conf[output_action] = resolved_ports
-                    elif output_action == 'failover':
-                        failover = output_action_values
-                        test_config_condition(not isinstance(failover, dict), (
-                            'failover is not a dictionary'))
-                        resolved_action_conf[output_action] = {}
-                        for failover_name, failover_values in list(failover.items()):
-                            if failover_name == 'ports':
-                                resolved_failover_values = resolve_port_numbers(failover_values)
-                                test_config_condition(len(resolved_failover_values) != len(failover_values), (
-                                    'failover port(s) not defined in DP: %s' % self.name))
-                                failover_values = resolved_failover_values
-                            resolved_action_conf[output_action][failover_name] = failover_values
-                    else:
-                        resolved_action_conf[output_action] = output_action_values
-                if resolved_action_conf:
-                    return resolved_action_conf
-                return None
-
-            def resolve_noop(_acl, action_conf):
-                return action_conf
-
-            action_resolvers = {
-                'meter': resolve_meter,
-                'mirror': resolve_mirror,
-                'output': resolve_output,
-                'allow': resolve_noop,
-                'force_port_vlan': resolve_noop,
-            }
-
-            def build_acl(acl, vid=None):
-                """Check that ACL can be built from config and mark mirror destinations."""
-                if acl.rules:
-                    try:
-                        ofmsgs = valve_acl.build_acl_ofmsgs(
-                            [acl], self.wildcard_table,
-                            valve_of.goto_table(self.wildcard_table),
-                            valve_of.goto_table(self.wildcard_table),
-                            2**16-1, self.meters, acl.exact_match,
-                            vlan_vid=vid)
-                        test_config_condition(not ofmsgs, 'of messages is empty')
-                        for ofmsg in ofmsgs:
-                            ofmsg.datapath = NullRyuDatapath()
-                            ofmsg.set_xid(0)
-                            ofmsg.serialize()
-                    except (AddrFormatError, KeyError, ValueError) as err:
-                        raise InvalidConfigError(err)
-                    for port_no in mirror_destinations:
-                        port = self.ports[port_no]
-                        port.output_only = True
-
-            for rule_conf in acl.rules:
-                for attrib, attrib_value in list(rule_conf.items()):
-                    if attrib == 'actions':
-                        resolved_actions = {}
-                        test_config_condition(not isinstance(attrib_value, dict), (
-                            'attrib_value is not a dictionary'))
-                        for action_name, action_conf in list(attrib_value.items()):
-                            resolved_action_conf = action_resolvers[action_name](
-                                acl, action_conf)
-                            test_config_condition(resolved_action_conf is None, (
-                                'cannot resolve ACL rule %s' % rule_conf))
-                            resolved_actions[action_name] = resolved_action_conf
-                        rule_conf[attrib] = resolved_actions
-
-            build_acl(acl, vid=1)
+            for port_no in acl.get_mirror_destinations():
+                port = self.ports[port_no]
+                port.output_only = True
+            return acl.build(vid, self.meters)
 
         def verify_acl_exact_match(acls):
             for acl in acls:
                 test_config_condition(acl.exact_match != acls[0].exact_match, (
                     'ACLs when used together must have consistent exact_match'))
+            return acls[0].exact_match
 
-        def resolve_acls():
+        def resolve_acls(valve_cl):
             """Resolve config references in ACLs."""
             # TODO: move this config validation to ACL object.
+            port_acl_enabled = valve_cl.STATIC_TABLE_IDS
+            port_acl_matches = {}
+            port_acl_set_fields = set()
+            port_acl_exact_match = False
+            port_acl_meter = False
+            vlan_acl_matches = {}
+            vlan_acl_exact_match = False
+            vlan_acl_set_fields = set()
+            vlan_acl_meter = False
+
+            def merge_matches(matches, new_matches):
+                for field, has_mask in list(new_matches.items()):
+                    if has_mask or field not in matches:
+                        matches[field] = has_mask
 
             for vlan in list(self.vlans.values()):
                 if vlan.acls_in:
                     acls = []
                     for acl in vlan.acls_in:
-                        resolve_acl(acl)
+                        matches, set_fields, meter = resolve_acl(acl, vlan.vid)
+                        merge_matches(vlan_acl_matches, matches)
+                        vlan_acl_set_fields = vlan_acl_set_fields.union(set_fields)
+                        if meter:
+                            vlan_acl_meter = True
                         acls.append(self.acls[acl])
                     vlan.acls_in = acls
-                    verify_acl_exact_match(acls)
+                    vlan_acl_exact_match = verify_acl_exact_match(acls)
             for port in list(self.ports.values()):
                 if port.acls_in:
                     test_config_condition(self.dp_acls, (
                         'dataplane ACLs cannot be used with port ACLs.'))
                     acls = []
                     for acl in port.acls_in:
-                        resolve_acl(acl)
+                        matches, set_fields, meter = resolve_acl(acl, None)
+                        merge_matches(port_acl_matches, matches)
+                        port_acl_set_fields = port_acl_set_fields.union(set_fields)
+                        if meter:
+                            port_acl_meter = True
                         acls.append(self.acls[acl])
                     port.acls_in = acls
-                    verify_acl_exact_match(acls)
+                    port_acl_exact_match = verify_acl_exact_match(acls)
+                    port_acl_enabled = True
             if self.dp_acls:
                 acls = []
                 for acl in self.acls:
-                    resolve_acl(acl)
+                    matches, set_fields, meter = resolve_acl(acl, None)
+                    merge_matches(port_acl_matches, matches)
+                    port_acl_set_fields = port_acl_set_fields.union(set_fields)
+                    if meter:
+                        port_acl_meter = True
                     acls.append(self.acls[acl])
                 self.dp_acls = acls
+                port_acl_enabled = True
+            if port_acl_enabled:
+                port_acl_matches.update({'in_port': False})
+            port_acl_matches = {(field, mask) for field, mask in list(port_acl_matches.items())}
+            vlan_acl_matches = {(field, mask) for field, mask in list(vlan_acl_matches.items())}
+
+            # TODO: skip port_acl table if not configured.
+            # TODO: dynamically configure output attribue
+            override_table_config = {
+                'port_acl': ValveTableConfig(
+                    'port_acl',
+                    exact_match=port_acl_exact_match,
+                    meter=port_acl_meter,
+                    output=True,
+                    match_types=port_acl_matches,
+                    set_fields=tuple(port_acl_set_fields)),
+                'vlan_acl': ValveTableConfig(
+                    'vlan_acl',
+                    exact_match=vlan_acl_exact_match,
+                    meter=vlan_acl_meter,
+                    output=True,
+                    match_types=vlan_acl_matches,
+                    set_fields=tuple(vlan_acl_set_fields)),
+            }
+            return override_table_config
 
         def resolve_vlan_names_in_routers():
             """Resolve VLAN references in routers."""
@@ -691,20 +769,53 @@ configuration.
                     dp_router = copy.copy(router)
                     dp_router.vlans = vlans
                     dp_routers[router_name] = dp_router
+                vips = set()
+                for vlan in vlans:
+                    for vip in vlan.faucet_vips:
+                        if vip.ip.is_link_local:
+                            continue
+                        vips.add(vip)
+                for vip in vips:
+                    for other_vip in vips - set([vip]):
+                        test_config_condition(
+                            vip.ip in other_vip.network,
+                            'VIPs %s and %s overlap in router %s' % (
+                                vip, other_vip, router_name))
             self.routers = dp_routers
 
         test_config_condition(not self.vlans, 'no VLANs referenced by interfaces in %s' % self.name)
+        valve_cl = SUPPORTED_HARDWARE.get(self.hardware, None)
+        test_config_condition(
+            not valve_cl, 'hardware %s must be in %s' % (
+                self.hardware, list(SUPPORTED_HARDWARE.keys())))
 
         for dp in dps:
             dp_by_name[dp.name] = dp
         for vlan in list(self.vlans.values()):
             vlan_by_name[vlan.name] = vlan
+            if self.global_vlan:
+                test_config_condition(
+                    self.global_vlan == vlan.vid, 'VLAN %u is reserved by global_vlan' % vlan.vid)
 
         resolve_stack_dps()
         resolve_mirror_destinations()
         resolve_override_output_ports()
         resolve_vlan_names_in_routers()
-        resolve_acls()
+        override_table_config = resolve_acls(valve_cl)
+
+        # Only configure IP routing tables if enabled.
+        ipvs = set()
+        for vlan in list(self.vlans.values()):
+            ipvs = ipvs.union(vlan.ipvs())
+        for ipv in (4, 6):
+            if ipv not in ipvs:
+                table_name = 'ipv%u_fib' % ipv
+                override_table_config[table_name] = ValveTableConfig(table_name)
+        if not ipvs:
+            override_table_config['vip'] = ValveTableConfig('vip')
+
+        vlan_port_factor = len(self.vlans) * len(self.ports)
+        self._configure_tables(override_table_config, valve_cl, vlan_port_factor)
 
         bgp_vlans = self.bgp_vlans()
         if bgp_vlans:
@@ -712,9 +823,9 @@ configuration.
                 vlan_dps = [dp for dp in dps if vlan.vid in dp.vlans]
                 test_config_condition(len(vlan_dps) != 1, (
                     'DPs %s sharing a BGP speaker VLAN is unsupported'))
-            router_ids = set([vlan.bgp_routerid for vlan in bgp_vlans])
+            router_ids = {vlan.bgp_routerid for vlan in bgp_vlans}
             test_config_condition(len(router_ids) != 1, 'BGP router IDs must all be the same')
-            bgp_ports = set([vlan.bgp_port for vlan in bgp_vlans])
+            bgp_ports = {vlan.bgp_port for vlan in bgp_vlans}
             test_config_condition(len(bgp_ports) != 1, 'BGP ports must all be the same')
             for vlan in bgp_vlans:
                 test_config_condition(vlan.bgp_server_addresses != (
@@ -739,11 +850,11 @@ configuration.
 
     def bgp_vlans(self):
         """Return list of VLANs with BGP enabled."""
-        return [vlan for vlan in list(self.vlans.values()) if vlan.bgp_as]
+        return tuple([vlan for vlan in list(self.vlans.values()) if vlan.bgp_as])
 
     def dot1x_ports(self):
         """Return list of ports with 802.1x enabled."""
-        return [port for port in list(self.ports.values()) if port.dot1x]
+        return tuple([port for port in list(self.ports.values()) if port.dot1x])
 
     def to_conf(self):
         """Return DP config as dict."""
@@ -769,18 +880,16 @@ configuration.
 
     def get_config_dict(self):
         """Return DP config as a dict for API call."""
-        if self.name:
-            vlans_dict = {}
-            for vlan in list(self.vlans.values()):
-                vlans_dict[vlan.name] = vlan.to_conf()
-            acls_dict = {}
-            for acl_id, acl in list(self.acls.items()):
-                acls_dict[acl_id] = acl.to_conf()
-            return {
-                'dps': {self.name: self.to_conf()},
-                'vlans': vlans_dict,
-                'acls': acls_dict}
-        return {}
+        vlans_dict = {}
+        for vlan in list(self.vlans.values()):
+            vlans_dict[vlan.name] = vlan.to_conf()
+        acls_dict = {}
+        for acl_id, acl in list(self.acls.items()):
+            acls_dict[acl_id] = acl.to_conf()
+        return {
+            'dps': {self.name: self.to_conf()},
+            'vlans': vlans_dict,
+            'acls': acls_dict}
 
     def _get_acl_config_changes(self, logger, new_dp):
         """Detect any config changes to ACLs.
@@ -926,8 +1035,14 @@ configuration.
                 changed_vlans (set): changed/added VLAN IDs.
                 all_ports_changed (bool): True if all ports changed.
         """
+        def _table_configs(dp):
+            return frozenset([
+                table.table_config for table in list(dp.tables.values())])
+
         if self.ignore_subconf(new_dp):
             logger.info('DP base level config changed - requires cold start')
+        elif _table_configs(self) != _table_configs(new_dp):
+            logger.info('pipeline table config change - requires cold start')
         elif new_dp.routers != self.routers:
             logger.info('DP routers config changed - requires cold start')
         else:
