@@ -21,7 +21,7 @@ from faucet import valve_of
 from faucet import valve_packet
 
 
-class ValveFloodManager(object):
+class ValveFloodManager:
     """Implement dataplane based flooding for standalone dataplanes."""
 
     # Enumerate possible eth_dst flood destinations.
@@ -51,7 +51,7 @@ class ValveFloodManager(object):
     @staticmethod
     def _vlan_all_ports(vlan, exclude_unicast):
         """Return list of all ports that should be flooded to on a VLAN."""
-        return vlan.flood_ports(vlan.get_ports(), exclude_unicast)
+        return list(vlan.flood_ports(vlan.get_ports(), exclude_unicast))
 
     @staticmethod
     def _build_flood_local_rule_actions(vlan, exclude_unicast, in_port):
@@ -133,10 +133,11 @@ class ValveFloodManager(object):
     def _build_multiout_flood_rules(self, vlan, command):
         """Build flooding rules for a VLAN without using groups."""
         flood_priority = self.flood_priority
-        mirror_acts = set()
+        mirror_acts = []
         for mirrored_port in vlan.mirrored_ports():
             for act in mirrored_port.mirror_actions():
-                mirror_acts.add(act)
+                mirror_acts.append(act)
+        mirror_acts = valve_of.dedupe_output_port_acts(mirror_acts)
         ofmsgs = []
         for unicast_eth_dst, eth_dst, eth_dst_mask in self.FLOOD_DSTS:
             if unicast_eth_dst and not vlan.unicast_flood:
@@ -201,9 +202,9 @@ class ValveFloodManager(object):
             return self._build_group_flood_rules(vlan, modify, command)
         return self._build_multiout_flood_rules(vlan, command)
 
-    def update_stack_topo(self, event, dp, port=None): # pylint: disable=unused-argument
+    def update_stack_topo(self, event, dp, port=None): # pylint: disable=unused-argument,invalid-name
         """Update the stack topology. It has nothing to do for non-stacking DPs."""
-        return
+        pass
 
     @staticmethod
     def edge_learn_port(_other_valves, pkt_meta):
@@ -255,8 +256,12 @@ class ValveFloodStackManager(ValveFloodManager):
 
         If a standalone switch, then flood to local VLAN ports.
 
-        If a distributed switch, use selective flooding
-        (see the following example).
+        If a distributed switch where all switches are directly
+        connected to the root (star topology), edge switches flood locally
+        and to the root, and the root floods to the other edges.
+
+        If a non-star distributed switch topologies, use selective
+        flooding (see the following example).
 
                                Hosts
                                ||||
@@ -309,18 +314,32 @@ class ValveFloodStackManager(ValveFloodManager):
         4: 5(s)
         5: 1 2 3 4
         """
+        exclude_ports = []
+        dp_local_in_port = self._port_is_dp_local(in_port)
+        if not dp_local_in_port:
+            in_port_peer_dp = in_port.stack['dp']
+            exclude_ports = [
+                port for port in self.stack_ports
+                if port.stack and port.stack['dp'] == in_port_peer_dp]
         local_flood_actions = self._build_flood_local_rule_actions(
             vlan, exclude_unicast, in_port)
         away_flood_actions = valve_of.flood_tagged_port_outputs(
-            self.away_from_root_stack_ports, in_port)
+            self.away_from_root_stack_ports, in_port, exclude_ports=exclude_ports)
         toward_flood_actions = valve_of.flood_tagged_port_outputs(
             self.towards_root_stack_ports, in_port)
         flood_all_except_self = away_flood_actions + local_flood_actions
 
-        # If we're the root of a distributed switch..
+        # TODO: optimization for 2 layer stack - no need to reflect off root.
+        # We should generalize the edge switch case, too.
+        if self.stack.get('longest_path_to_root_len', None) == 2:
+            if self._dp_is_root():
+                return away_flood_actions + local_flood_actions
+            if dp_local_in_port:
+                return toward_flood_actions + local_flood_actions
+            return local_flood_actions
+
         if self._dp_is_root():
-            # If the input port was local, then flood local VLAN and stacks.
-            if self._port_is_dp_local(in_port):
+            if dp_local_in_port:
                 return flood_all_except_self
             # If input port non-local, then flood outward again
             return [valve_of.output_in_port()] + flood_all_except_self
@@ -358,6 +377,8 @@ class ValveFloodStackManager(ValveFloodManager):
             command = valve_of.ofp.OFPFC_MODIFY_STRICT
         # TODO: group tables for stacking
         ofmsgs = self._build_multiout_flood_rules(vlan, command)
+        if self._dp_is_root():
+            return ofmsgs
         # Because stacking uses reflected broadcasts from the root,
         # don't try to learn broadcast sources from stacking ports.
         for port in self.stack_ports:
@@ -402,8 +423,7 @@ class ValveFloodStackManager(ValveFloodManager):
             return False
         return True
 
-    @staticmethod
-    def _edge_dp_for_host(other_valves, pkt_meta):
+    def _edge_dp_for_host(self, other_valves, pkt_meta):
         """Simple distributed unicast learning.
 
         Args:
@@ -426,9 +446,13 @@ class ValveFloodStackManager(ValveFloodManager):
         # TODO: stacking handles failure of redundant links between DPs,
         # but not failure of an entire DP (should be able to find
         # shortest path via alternate DP).
+        if pkt_meta.port.stack:
+            peer_dp = pkt_meta.port.stack['dp']
+            if peer_dp.is_stack_edge() or peer_dp.is_stack_root():
+                return peer_dp
+        stacked_valves = [valve for valve in other_valves if valve.dp.stack is not None]
         eth_src = pkt_meta.eth_src
         vlan_vid = pkt_meta.vlan.vid
-        stacked_valves = [valve for valve in other_valves if valve.dp.stack is not None]
         for other_valve in stacked_valves:
             if vlan_vid in other_valve.dp.vlans:
                 other_dp_vlan = other_valve.dp.vlans[vlan_vid]
@@ -441,17 +465,22 @@ class ValveFloodStackManager(ValveFloodManager):
     def update_stack_topo(self, event, dp, port=None):
         """Update the stack topo according to the event."""
 
-        def _stack_topo_up_dp(dp): # pylint: disable=unused-argument
-            pass
+        def _stack_topo_up_dp(_dp): # pylint: disable=invalid-name
+            for port in [port for port in _dp.stack_ports]:
+                if port.is_stack_up():
+                    _stack_topo_up_port(_dp, port)
+                else:
+                    _stack_topo_down_port(_dp, port)
 
-        def _stack_topo_down_dp(dp): # pylint: disable=unused-argument
-            pass
+        def _stack_topo_down_dp(_dp): # pylint: disable=invalid-name
+            for port in [port for port in _dp.stack_ports]:
+                _stack_topo_down_port(_dp, port)
 
-        def _stack_topo_up_port(dp, port): # pylint: disable=unused-argument
-            pass
+        def _stack_topo_up_port(_dp, _port): # pylint: disable=invalid-name
+            _dp.add_stack_link(self.stack['graph'], _dp, _port)
 
-        def _stack_topo_down_port(dp, port): # pylint: disable=unused-argument
-            pass
+        def _stack_topo_down_port(_dp, _port): # pylint: disable=invalid-name
+            _dp.remove_stack_link(self.stack['graph'], _dp, _port)
 
         if port:
             if event:
@@ -463,6 +492,7 @@ class ValveFloodStackManager(ValveFloodManager):
                 _stack_topo_up_dp(dp)
             else:
                 _stack_topo_down_dp(dp)
+        return True
 
     def edge_learn_port(self, other_valves, pkt_meta):
         """Possibly learn a host on a port.
