@@ -22,7 +22,7 @@ import random
 from faucet import valve_of
 
 
-class ValveHostManager(object):
+class ValveHostManager:
     """Manage host learning on VLANs."""
 
     # don't update host cache more often than this many seconds
@@ -147,29 +147,9 @@ class ValveHostManager(object):
                 priority=(self.host_priority - 2)))
         else:
             # Delete any existing entries for MAC.
-            # TODO: for LAGs, don't delete entries in the same LAG.
             if delete_existing:
                 ofmsgs.extend(self.delete_host_from_vlan(eth_src, vlan))
 
-        # Associate this MAC with source port.
-        src_match = self.eth_src_table.match(
-            in_port=port.number, vlan=vlan, eth_src=eth_src)
-        if port.override_output_port:
-            ofmsgs.append(self.eth_src_table.flowmod(
-                match=src_match,
-                priority=(self.host_priority - 1),
-                inst=[valve_of.apply_actions([
-                    valve_of.output_port(port.override_output_port.number)])],
-                hard_timeout=src_rule_hard_timeout,
-                idle_timeout=src_rule_idle_timeout))
-        else:
-            ofmsgs.append(self.eth_src_table.flowmod(
-                match=src_match,
-                priority=(self.host_priority - 1),
-                inst=[valve_of.goto_table(self.eth_dst_table)],
-                hard_timeout=src_rule_hard_timeout,
-                idle_timeout=src_rule_idle_timeout))
-
         # Output packets for this MAC to specified port.
         ofmsgs.append(self.eth_dst_table.flowmod(
             self.eth_dst_table.match(vlan=vlan, eth_dst=eth_src),
@@ -177,15 +157,32 @@ class ValveHostManager(object):
             inst=[valve_of.apply_actions(vlan.output_port(port))],
             idle_timeout=dst_rule_idle_timeout))
 
+        # Associate this MAC with source port.
+        src_match = self.eth_src_table.match(
+            in_port=port.number, vlan=vlan, eth_src=eth_src)
+        src_priority = self.host_priority - 1
+        inst = valve_of.goto_table(self.eth_dst_table)
+
+        if port.override_output_port:
+            inst = valve_of.apply_actions([
+                valve_of.output_port(port.override_output_port.number)])
+        ofmsgs.append(self.eth_src_table.flowmod(
+            match=src_match,
+            priority=src_priority,
+            inst=[inst],
+            hard_timeout=src_rule_hard_timeout,
+            idle_timeout=src_rule_idle_timeout))
+
         # If port is in hairpin mode, install a special rule
         # that outputs packets destined to this MAC back out the same
         # port they came in (e.g. multiple hosts on same WiFi AP,
         # and FAUCET is switching between them on the same port).
         if port.hairpin:
-            ofmsgs.append(self.eth_dst_table.flowmod(
-                self.eth_dst_table.match(in_port=port.number, vlan=vlan, eth_dst=eth_src),
+            ofmsgs.append(self.eth_src_table.flowmod(
+                self.eth_src_table.match(in_port=port.number, vlan=vlan, eth_dst=eth_src),
                 priority=(self.host_priority + 1),
                 inst=[valve_of.apply_actions(vlan.output_port(port, hairpin=True))],
+                hard_timeout=src_rule_hard_timeout,
                 idle_timeout=dst_rule_idle_timeout))
 
         return ofmsgs
@@ -209,8 +206,11 @@ class ValveHostManager(object):
             cache_age = now - entry.cache_time
             cache_port = entry.port
 
-        if cache_port == port:
-            # skip delete if host didn't change ports.
+        same_lag = (
+            cache_port is not None and
+            cache_port.lacp and port.lacp and cache_port.lacp == port.lacp)
+        if cache_port == port or same_lag:
+            # skip delete if host didn't change ports or on same LAG.
             delete_existing = False
             # if we very very recently learned this host, don't do anything.
             if cache_age < self.CACHE_UPDATE_GUARD_TIME:
