@@ -21,7 +21,8 @@ import ipaddress
 
 from ryu.lib import mac
 from ryu.lib import ofctl_v1_3 as ofctl
-from ryu.lib.ofctl_utils import str_to_int, to_match_ip, to_match_masked_int, to_match_eth, to_match_vid, OFCtlUtil
+from ryu.lib.ofctl_utils import (
+    str_to_int, to_match_ip, to_match_masked_int, to_match_eth, to_match_vid, OFCtlUtil)
 from ryu.ofproto import ether
 from ryu.ofproto import inet
 from ryu.ofproto import ofproto_v1_3 as ofp
@@ -37,6 +38,7 @@ ROUTE_GROUP_OFFSET = VLAN_GROUP_OFFSET * 2
 OFP_VERSIONS = [ofp.OFP_VERSION]
 OFP_IN_PORT = ofp.OFPP_IN_PORT
 MAX_PACKET_IN_BYTES = 128
+ECTP_ETH_TYPE = 0x9000
 
 
 def ignore_port(port_num):
@@ -56,7 +58,7 @@ def ignore_port(port_num):
 
 def port_status_from_state(state):
     """Return True if OFPPS_LINK_DOWN is not set."""
-    return not (state & ofp.OFPPS_LINK_DOWN)
+    return not state & ofp.OFPPS_LINK_DOWN
 
 
 def is_table_features_req(ofmsg):
@@ -174,6 +176,33 @@ def is_meteradd(ofmsg):
     return False
 
 
+def is_apply_actions(instruction):
+    """Return True if an apply action.
+
+    Args:
+        instruction: OpenFlow instruction.
+    Returns:
+        bool: True if an apply action.
+    """
+    return (isinstance(instruction, parser.OFPInstructionActions) and
+            instruction.type == ofp.OFPIT_APPLY_ACTIONS)
+
+
+def is_meter(instruction):
+    """Return True if a meter.
+
+    Args:
+        instruction: OpenFlow instruction.
+    Returns:
+        bool: True if a meter.
+    """
+    return isinstance(instruction, parser.OFPInstructionMeter)
+
+
+def is_set_field(action):
+    return isinstance(action, parser.OFPActionSetField)
+
+
 def apply_meter(meter_id):
     """Return instruction to apply a meter."""
     return parser.OFPInstructionMeter(meter_id, ofp.OFPIT_METER)
@@ -311,6 +340,18 @@ def output_port(port_num, max_len=0):
     return parser.OFPActionOutput(port_num, max_len=max_len)
 
 
+def dedupe_output_port_acts(output_port_acts):
+    """Deduplicate parser.OFPActionOutputs (because Ryu doesn't define __eq__).
+
+    Args:
+        list of ryu.ofproto.ofproto_v1_3_parser.OFPActionOutput: output to port actions.
+    Returns:
+        list of ryu.ofproto.ofproto_v1_3_parser.OFPActionOutput: output to port actions.
+    """
+    output_ports = {output_port_act.port for output_port_act in output_port_acts}
+    return [output_port(port) for port in sorted(output_ports)]
+
+
 def output_in_port():
     """Return OpenFlow action to output out input port.
 
@@ -568,7 +609,7 @@ def meterdel(datapath=None, meter_id=ofp.OFPM_ALL):
 def meteradd(meter_conf):
     """Add a meter based on YAML configuration."""
 
-    class NoopDP(object):
+    class NoopDP:
         """Fake DP to be able to use ofctl to parse meter config."""
 
         id = 0
@@ -717,13 +758,15 @@ def faucet_config(datapath=None):
     return parser.OFPSetConfig(datapath, ofp.OFPC_FRAG_NORMAL, 0)
 
 
-def faucet_async(datapath=None, notify_flow_removed=False, packet_in=True):
+def faucet_async(datapath=None, notify_flow_removed=False, packet_in=True, port_status=True):
     """Return async message config for FAUCET/Gauge"""
     packet_in_mask = 0
     if packet_in:
         packet_in_mask = 1 << ofp.OFPR_ACTION
-    port_status_mask = (
-        1 << ofp.OFPPR_ADD | 1 << ofp.OFPPR_DELETE | 1 << ofp.OFPPR_MODIFY)
+    port_status_mask = 0
+    if port_status:
+        port_status_mask = (
+            1 << ofp.OFPPR_ADD | 1 << ofp.OFPPR_DELETE | 1 << ofp.OFPPR_MODIFY)
     flow_removed_mask = 0
     if notify_flow_removed:
         flow_removed_mask = (
