@@ -22,13 +22,14 @@ import socket
 import struct
 
 from ryu.lib import addrconv
+from ryu.lib.mac import BROADCAST, is_multicast, haddr_to_bin
 from ryu.lib.packet import (
     arp, bpdu, ethernet,
     icmp, icmpv6, ipv4, ipv6,
     lldp, slow, packet, vlan)
 from ryu.lib.packet.stream_parser import StreamParser
 
-from faucet.valve_util import btos
+from faucet import valve_util
 from faucet import valve_of
 
 FAUCET_MAC = '0e:00:00:00:00:01' # Default FAUCET MAC address
@@ -46,14 +47,24 @@ LLDP_MAC_NEAREST_BRIDGE = lldp.LLDP_MAC_NEAREST_BRIDGE
 CISCO_SPANNING_GROUP_ADDRESS = '01:00:0c:cc:cc:cd'
 IPV6_ALL_NODES_MCAST = '33:33:00:00:00:01'
 IPV6_ALL_ROUTERS_MCAST = '33:33:00:00:00:02'
-IPV6_LINK_LOCAL = ipaddress.IPv6Network(btos('fe80::/10'))
-IPV6_ALL_NODES = ipaddress.IPv6Address(btos('ff02::1'))
+IPV6_ALL_NODES = ipaddress.IPv6Address(valve_util.btos('ff02::1'))
 IPV6_MAX_HOP_LIM = 255
 
 LLDP_FAUCET_DP_ID = 1
 LLDP_FAUCET_STACK_STATE = 2
 
 
+def int_from_mac(mac):
+    int_hi, int_lo = [int(i, 16) for i in mac.split(':')[-2:]]
+    return (int_hi << 8) + int_lo
+
+
+def int_in_mac(mac, to_int):
+    int_mac = mac.split(':')[:4] + [
+        '%x' % (to_int >> 8), '%x' % (to_int & 0xff)]
+    return ':'.join(int_mac)
+
+
 def ipv4_parseable(ip_header_data):
     """Return True if an IPv4 packet we could parse."""
     # TODO: python library parsers are fragile
@@ -94,17 +105,6 @@ def parse_eth_pkt(pkt):
     return pkt.get_protocol(ethernet.ethernet)
 
 
-def parse_vlan_pkt(pkt):
-    """Return parsed VLAN header.
-
-    Args:
-        pkt (ryu.lib.packet.packet): packet received from dataplane.
-    Returns:
-        ryu.lib.packet.vlan: VLAN header.
-    """
-    return pkt.get_protocol(vlan.vlan)
-
-
 def parse_lacp_pkt(pkt):
     """Return parsed LACP packet.
 
@@ -127,7 +127,7 @@ def parse_lldp(pkt):
     return pkt.get_protocol(lldp.lldp)
 
 
-def parse_packet_in_pkt(data, max_len):
+def parse_packet_in_pkt(data, max_len, eth_pkt=None, vlan_pkt=None):
     """Parse a packet received via packet in from the dataplane.
 
     Args:
@@ -140,7 +140,6 @@ def parse_packet_in_pkt(data, max_len):
         int: VLAN VID (or None if no VLAN)
     """
     pkt = None
-    eth_pkt = None
     eth_type = None
     vlan_vid = None
 
@@ -150,21 +149,22 @@ def parse_packet_in_pkt(data, max_len):
     try:
         # Packet may or may not have a VLAN tag - whether it is user
         # traffic, or control like LACP/LLDP.
-        pkt = packet.Packet(data[:ETH_HEADER_SIZE])
-        eth_pkt = parse_eth_pkt(pkt)
-        eth_type = eth_pkt.ethertype
-        if eth_type == valve_of.ether.ETH_TYPE_8021Q:
-            pkt = packet.Packet(data[:ETH_VLAN_HEADER_SIZE])
-            vlan_pkt = parse_vlan_pkt(pkt)
-            if vlan_pkt:
-                vlan_vid = vlan_pkt.vid
-                eth_type = vlan_pkt.ethertype
+        if vlan_pkt is None:
+            if eth_pkt is None:
+                pkt = packet.Packet(data[:ETH_HEADER_SIZE])
+                eth_pkt = parse_eth_pkt(pkt)
+            eth_type = eth_pkt.ethertype
+            if eth_type == valve_of.ether.ETH_TYPE_8021Q:
+                pkt, vlan_pkt = packet.Packet(data[:ETH_VLAN_HEADER_SIZE])
+        if vlan_pkt:
+            vlan_vid = vlan_pkt.vid
+            eth_type = vlan_pkt.ethertype
         if len(data) > ETH_VLAN_HEADER_SIZE:
             pkt = packet.Packet(data)
     except (AttributeError, AssertionError, StreamParser.TooSmallException):
         pass
 
-    return (pkt, eth_pkt, eth_type, vlan_vid)
+    return (pkt, eth_pkt, eth_type, vlan_pkt, vlan_vid)
 
 
 def mac_addr_is_unicast(mac_addr):
@@ -175,8 +175,10 @@ def mac_addr_is_unicast(mac_addr):
     Returns:
         bool: True if a unicast Ethernet address.
     """
-    msb = mac_addr.split(':')[0]
-    return msb[-1] in '02468aAcCeE'
+    mac_bin = haddr_to_bin(mac_addr)
+    if mac_bin == BROADCAST:
+        return False
+    return not is_multicast(mac_bin)
 
 
 def build_pkt_header(vid, eth_src, eth_dst, dl_type):
@@ -275,6 +277,54 @@ def faucet_lldp_stack_state_tlvs(dp, port):
     return tlvs
 
 
+def tlvs_by_type(tlvs, tlv_type):
+    """Return list of TLVs with matching type."""
+    return [tlv for tlv in tlvs if tlv.tlv_type == tlv_type]
+
+
+def tlvs_by_subtype(tlvs, subtype):
+    """Return list of TLVs with matching type."""
+    return [tlv for tlv in tlvs if tlv.subtype == subtype]
+
+
+def tlv_cast(tlvs, tlv_attr, cast_func):
+    """Return cast'd attribute of first TLV or None."""
+    tlv_val = None
+    if tlvs:
+        try:
+            tlv_val = cast_func(getattr(tlvs[0], tlv_attr))
+        except (AttributeError, ValueError, TypeError):
+            pass
+    return tlv_val
+
+
+def faucet_tlvs(lldp_pkt, faucet_dp_mac):
+    """Return list of TLVs with FAUCET OUI."""
+    return [tlv for tlv in tlvs_by_type(
+        lldp_pkt.tlvs, lldp.LLDP_TLV_ORGANIZATIONALLY_SPECIFIC)
+            if tlv.oui == faucet_oui(faucet_dp_mac)]
+
+
+def parse_faucet_lldp(lldp_pkt, faucet_dp_mac):
+    """Parse and return FAUCET TLVs from LLDP packet."""
+    remote_dp_id = None
+    remote_dp_name = None
+    remote_port_id = None
+    remote_port_state = None
+
+    tlvs = faucet_tlvs(lldp_pkt, faucet_dp_mac)
+    if tlvs:
+        dp_id_tlvs = tlvs_by_subtype(tlvs, LLDP_FAUCET_DP_ID)
+        dp_name_tlvs = tlvs_by_type(lldp_pkt.tlvs, lldp.LLDP_TLV_SYSTEM_NAME)
+        port_id_tlvs = tlvs_by_type(lldp_pkt.tlvs, lldp.LLDP_TLV_PORT_ID)
+        port_state_tlvs = tlvs_by_subtype(tlvs, LLDP_FAUCET_STACK_STATE)
+        remote_dp_id = tlv_cast(dp_id_tlvs, 'info', int)
+        remote_port_id = tlv_cast(port_id_tlvs, 'port_id', int)
+        remote_port_state = tlv_cast(port_state_tlvs, 'info', int)
+        remote_dp_name = tlv_cast(dp_name_tlvs, 'system_name', valve_util.utf8_decode)
+    return (remote_dp_id, remote_dp_name, remote_port_id, remote_port_state)
+
+
 def lacp_reqreply(eth_src,
                   actor_system, actor_key, actor_port,
                   partner_system, partner_key, partner_port,
@@ -442,7 +492,7 @@ def ipv6_solicited_node_from_ucast(ucast):
     Returns:
        ipaddress.IPv6Address: IPv6 solicited node multicast address.
     """
-    link_mcast_prefix = ipaddress.ip_interface(btos('ff02::1:ff00:0/104'))
+    link_mcast_prefix = ipaddress.ip_interface(valve_util.btos('ff02::1:ff00:0/104'))
     mcast_bytes = link_mcast_prefix.packed[:13] + ucast.packed[-3:]
     link_mcast = ipaddress.IPv6Address(mcast_bytes)
     return link_mcast
@@ -583,9 +633,25 @@ def router_advert(vid, eth_src, eth_dst, src_ip, dst_ip,
     return pkt
 
 
-class PacketMeta(object):
+class PacketMeta:
     """Original, and parsed Ethernet packet metadata."""
 
+    __slots__ = [
+        'data',
+        'orig_len',
+        'pkt',
+        'eth_pkt',
+        'vlan_pkt',
+        'port',
+        'vlan',
+        'eth_src',
+        'eth_dst',
+        'eth_type',
+        'l3_pkt',
+        'l3_src',
+        'l3_dst',
+    ]
+
     ETH_TYPES_PARSERS = {
         valve_of.ether.ETH_TYPE_IP: (4, ipv4_parseable, ipv4.ipv4),
         valve_of.ether.ETH_TYPE_ARP: (None, None, arp.arp),
@@ -598,11 +664,13 @@ class PacketMeta(object):
         valve_of.ether.ETH_TYPE_IPV6: ETH_VLAN_HEADER_SIZE + IPV6_HEADER_SIZE,
     }
 
-    def __init__(self, data, orig_len, pkt, eth_pkt, port, valve_vlan, eth_src, eth_dst, eth_type):
+    def __init__(self, data, orig_len, pkt, eth_pkt, vlan_pkt, port, valve_vlan,
+                 eth_src, eth_dst, eth_type):
         self.data = data
         self.orig_len = orig_len
         self.pkt = pkt
         self.eth_pkt = eth_pkt
+        self.vlan_pkt = vlan_pkt
         self.port = port
         self.vlan = valve_vlan
         self.eth_src = eth_src
@@ -612,14 +680,23 @@ class PacketMeta(object):
         self.l3_src = None
         self.l3_dst = None
 
+    def log(self):
+        vlan_msg = ''
+        if self.vlan:
+            vlan_msg = 'VLAN %u' % self.vlan.vid
+        return '%s (L2 type 0x%4.4x, L3 src %s, L3 dst %s) %s %s' % (
+            self.eth_src, self.eth_type, self.l3_src, self.l3_dst,
+            self.port, vlan_msg)
+
     def reparse(self, max_len):
         """Reparse packet using data up to the specified maximum length."""
-        pkt, eth_pkt, eth_type, _ = parse_packet_in_pkt(
-            self.data, max_len)
+        pkt, eth_pkt, eth_type, vlan_pkt, _ = parse_packet_in_pkt(
+            self.data, max_len, eth_pkt=self.eth_pkt, vlan_pkt=self.vlan_pkt)
         if pkt is None or eth_type is None:
             return
         self.pkt = pkt
         self.eth_pkt = eth_pkt
+        self.vlan_pkt = vlan_pkt
 
     def reparse_all(self):
         """Reparse packet with all available data."""
@@ -640,7 +717,7 @@ class PacketMeta(object):
             if ip_ver is not None:
                 if ip_ver != self.ip_ver():
                     return
-                if self.vlan.minimum_ip_size_check:
+                if self.vlan is not None and self.vlan.minimum_ip_size_check:
                     if len(self.data) < header_size:
                         return
                 ip_header_data = self.data[ETH_VLAN_HEADER_SIZE:]
@@ -656,6 +733,8 @@ class PacketMeta(object):
                 elif hasattr(self.l3_pkt, 'src_ip'):
                     self.l3_src = self.l3_pkt.src_ip
                     self.l3_dst = self.l3_pkt.dst_ip
+                self.l3_src = ipaddress.ip_address(valve_util.btos(self.l3_src))
+                self.l3_dst = ipaddress.ip_address(valve_util.btos(self.l3_dst))
 
     def packet_complete(self):
         """True if we have the complete packet."""
