@@ -28,9 +28,23 @@ from faucet import valve_packet
 from faucet.valve_util import btos
 
 
-class NextHop(object):
+class AnonVLAN:
+
+    def __init__(self, vid):
+        self.vid = vid
+
+
+class NextHop:
     """Describes a directly connected (at layer 2) nexthop."""
 
+    __slots__ = [
+        'cache_time',
+        'eth_src',
+        'last_retry_time',
+        'resolve_retries',
+        'port',
+    ]
+
     def __init__(self, eth_src, port, now):
         self.eth_src = eth_src
         self.port = port
@@ -47,22 +61,41 @@ class NextHop(object):
         return self.resolve_retries >= max_fib_retries
 
 
-class ValveRouteManager(object):
+class ValveRouteManager:
     """Base class to implement RIB/FIB."""
 
+    __slots__ = [
+        'active',
+        'arp_neighbor_timeout',
+        'dec_ttl',
+        'eth_dst_table',
+        'eth_src_table',
+        'fib_table',
+        'flood_table',
+        'global_vlan',
+        'logger',
+        'max_host_fib_retry_count',
+        'max_hosts_per_resolve_cycle',
+        'max_resolve_backoff_time',
+        'proactive_learn',
+        'route_priority',
+        'routers',
+        'vip_table',
+    ]
+
     IPV = 0
     ETH_TYPE = None
     ICMP_TYPE = None
     MAX_LEN = valve_of.MAX_PACKET_IN_BYTES
     CONTROL_ETH_TYPES = () # type: ignore
-    active = False
 
-    def __init__(self, logger, arp_neighbor_timeout,
+    def __init__(self, logger, global_vlan, arp_neighbor_timeout,
                  max_hosts_per_resolve_cycle, max_host_fib_retry_count,
                  max_resolve_backoff_time, proactive_learn, dec_ttl,
                  fib_table, vip_table, eth_src_table, eth_dst_table, flood_table,
-                 route_priority, routers, use_group_table, groups):
+                 route_priority, routers):
         self.logger = logger
+        self.global_vlan = AnonVLAN(global_vlan)
         self.arp_neighbor_timeout = arp_neighbor_timeout
         self.max_hosts_per_resolve_cycle = max_hosts_per_resolve_cycle
         self.max_host_fib_retry_count = max_host_fib_retry_count
@@ -76,8 +109,9 @@ class ValveRouteManager(object):
         self.flood_table = flood_table
         self.route_priority = route_priority
         self.routers = routers
-        self.use_group_table = use_group_table
-        self.groups = groups
+        self.active = False
+        if self._global_routing():
+            self.logger.info('global routing enabled')
 
     def nexthop_dead(self, nexthop_cache_entry):
         return nexthop_cache_entry.dead(self.max_host_fib_retry_count)
@@ -104,10 +138,6 @@ class ValveRouteManager(object):
         nexthop_cache = self._vlan_nexthop_cache(vlan)
         del nexthop_cache[ip_gw]
 
-    def _group_id_from_ip_gw(self, vlan, resolved_ip_gw):
-        return self.groups.group_id_from_str(
-            ''.join((str(vlan), str(resolved_ip_gw))))
-
     def _nexthop_actions(self, eth_dst, vlan):
         ofmsgs = []
         if self.routers:
@@ -119,40 +149,63 @@ class ValveRouteManager(object):
             ofmsgs.append(valve_of.dec_ip_ttl())
         return ofmsgs
 
-    def _route_match(self, vlan, ip_dst):
-        return self.fib_table.match(vlan=vlan, eth_type=self.ETH_TYPE, nw_dst=ip_dst)
+    def _route_match(self, vlan, ip_dst=None):
+        if ip_dst is not None:
+            return self.fib_table.match(vlan=vlan, eth_type=self.ETH_TYPE, nw_dst=ip_dst)
+        return self.fib_table.match(vlan=vlan, eth_type=self.ETH_TYPE)
 
     def _route_priority(self, ip_dst):
         prefixlen = ipaddress.ip_network(ip_dst).prefixlen
         return self.route_priority + prefixlen
 
+    def _router_for_vlan(self, vlan):
+        if self.routers:
+            for router in list(self.routers.values()):
+                if vlan in router.vlans:
+                    return router
+        return None
+
     def _routed_vlans(self, vlan):
+        if self._global_routing():
+            return set([self.global_vlan])
         vlans = set([vlan])
         if self.routers:
             for router in list(self.routers.values()):
                 if vlan in router.vlans:
-                    for other_vlan in router.vlans:
-                        vlans.add(other_vlan)
+                    vlans = vlans.union(router.vlans)
         return vlans
 
+    def _stateful_gw(self, vlan, dst_ip):
+        return vlan.ip_dsts_for_ip_gw(dst_ip) or not dst_ip.is_link_local
+
+    def _global_routing(self):
+        return self.global_vlan.vid and self.routers and len(self.routers) == 1
+
     def _add_faucet_fib_to_vip(self, vlan, priority, faucet_vip, faucet_vip_host):
         learn_connected_priority = self.route_priority + faucet_vip.network.prefixlen
+        faucet_mac = vlan.faucet_mac
+        insts = [valve_of.goto_table(self.fib_table)]
+        if self._global_routing():
+            vlan_mac = valve_packet.int_in_mac(faucet_mac, vlan.vid)
+            insts = [valve_of.apply_actions([
+                valve_of.set_eth_dst(vlan_mac),
+                valve_of.set_vlan_vid(self.global_vlan.vid)])] + insts
         ofmsgs = []
         ofmsgs.append(self.eth_src_table.flowmod(
-            self.eth_src_table.match(eth_type=self.ETH_TYPE, eth_dst=vlan.faucet_mac, vlan=vlan),
+            self.eth_src_table.match(eth_type=self.ETH_TYPE, eth_dst=faucet_mac, vlan=vlan),
             priority=self.route_priority,
-            inst=[valve_of.goto_table(self.fib_table)]))
+            inst=insts))
+        routed_vlans = self._routed_vlans(vlan)
+        if self._global_routing():
+            vlan = self.global_vlan
         ofmsgs.append(self.fib_table.flowmod(
-            self.fib_table.match(eth_type=self.ETH_TYPE, vlan=vlan, nw_dst=faucet_vip_host),
+            self._route_match(vlan, faucet_vip_host),
             priority=priority,
             inst=[valve_of.goto_table(self.vip_table)]))
-        if self.proactive_learn:
-            for routed_vlan in self._routed_vlans(vlan):
+        if self.proactive_learn and not faucet_vip.ip.is_link_local:
+            for routed_vlan in routed_vlans:
                 ofmsgs.append(self.fib_table.flowmod(
-                    self.fib_table.match(
-                        eth_type=self.ETH_TYPE,
-                        vlan=routed_vlan,
-                        nw_dst=faucet_vip),
+                    self._route_match(routed_vlan, faucet_vip),
                     priority=learn_connected_priority,
                     inst=[valve_of.goto_table(self.vip_table)]))
             ofmsgs.append(self.vip_table.flowcontroller(
@@ -186,13 +239,10 @@ class ValveRouteManager(object):
             self.logger.info(
                 'Adding new route %s via %s (%s) on VLAN %u' % (
                     ip_dst, ip_gw, eth_dst, vlan.vid))
-        if self.use_group_table:
-            inst = [valve_of.apply_actions([valve_of.group_act(
-                group_id=self._group_id_from_ip_gw(vlan, ip_gw))])]
-        else:
-            inst = [valve_of.apply_actions(self._nexthop_actions(eth_dst, vlan)),
-                    valve_of.goto_table(self.eth_dst_table)]
-        for routed_vlan in self._routed_vlans(vlan):
+        inst = [valve_of.apply_actions(self._nexthop_actions(eth_dst, vlan)),
+                valve_of.goto_table(self.eth_dst_table)]
+        routed_vlans = self._routed_vlans(vlan)
+        for routed_vlan in routed_vlans:
             in_match = self._route_match(routed_vlan, ip_dst)
             ofmsgs.append(self.fib_table.flowmod(
                 in_match, priority=self._route_priority(ip_dst), inst=inst))
@@ -204,25 +254,6 @@ class ValveRouteManager(object):
         nexthop_cache[ip_gw] = nexthop
         return nexthop
 
-    def _nexthop_group_buckets(self, vlan, port, eth_src):
-        actions = self._nexthop_actions(eth_src, vlan)
-        actions.extend(vlan.output_port(port))
-        buckets = [valve_of.bucket(actions=actions)]
-        return buckets
-
-    def _update_nexthop_group(self, is_updated, resolved_ip_gw,
-                              vlan, port, eth_src):
-        group_id = self._group_id_from_ip_gw(vlan, resolved_ip_gw)
-        buckets = self._nexthop_group_buckets(vlan, port, eth_src)
-        nexthop_group = self.groups.get_entry(
-            group_id, buckets)
-        ofmsgs = []
-        if is_updated:
-            ofmsgs.append(nexthop_group.modify())
-        else:
-            ofmsgs.extend(nexthop_group.add())
-        return ofmsgs
-
     def _update_nexthop(self, now, vlan, port, eth_src, resolved_ip_gw):
         """Update routes where nexthop is newly resolved or changed.
 
@@ -240,12 +271,6 @@ class ValveRouteManager(object):
 
         if cached_eth_dst != eth_src:
             is_updated = cached_eth_dst is not None
-
-            if self.use_group_table:
-                ofmsgs.extend(
-                    self._update_nexthop_group(
-                        is_updated, resolved_ip_gw,
-                        vlan, port, eth_src))
             for ip_dst in vlan.ip_dsts_for_ip_gw(resolved_ip_gw):
                 ofmsgs.extend(self._add_resolved_route(
                     vlan, resolved_ip_gw, ip_dst, eth_src, is_updated))
@@ -259,17 +284,11 @@ class ValveRouteManager(object):
         Args:
             vlan (vlan): VLAN containing this RIB/FIB.
         Returns:
-            list: tuple, gateway, controller IP in same subnet.
+            tuple: frozenset, route gateways, host routes.
         """
-        host_ip_gws = []
-        route_ip_gws = []
-        for ip_gw in vlan.all_ip_gws(self.IPV):
-            for faucet_vip in vlan.faucet_vips_by_ipv(self.IPV):
-                if ip_gw in faucet_vip.network:
-                    if self._is_host_fib_route(vlan, ip_gw):
-                        host_ip_gws.append((ip_gw, faucet_vip))
-                    else:
-                        route_ip_gws.append((ip_gw, faucet_vip))
+        all_ip_gws = vlan.all_ip_gws(self.IPV)
+        host_ip_gws = frozenset([ip_gw for ip_gw in all_ip_gws if vlan.is_host_fib_route(ip_gw)])
+        route_ip_gws = frozenset(all_ip_gws - host_ip_gws)
         return (route_ip_gws, host_ip_gws)
 
     def _retry_backoff(self, now, resolve_retries, last_retry_time):
@@ -287,18 +306,19 @@ class ValveRouteManager(object):
            ip_gws (list): tuple, IP gateway and controller IP in same subnet.
            now (float): seconds since epoch.
         Returns:
-           list: tuple, gateway, controller IP in same subnet, cache entry.
+           list: prioritized list of gateways
         """
         ip_gws_never_tried = []
         ip_gws_with_retry_time = []
         not_fresh_nexthops = [
-            (ip_gw, faucet_vip) for ip_gw, faucet_vip in ip_gws if not self._nexthop_fresh(vlan, ip_gw, now)]
-        for ip_gw, faucet_vip in not_fresh_nexthops:
+            ip_gw for ip_gw in ip_gws
+            if not self._nexthop_fresh(vlan, ip_gw, now)]
+        for ip_gw in not_fresh_nexthops:
             nexthop_cache_entry = self._vlan_nexthop_cache_entry(vlan, ip_gw)
             if nexthop_cache_entry is None:
                 nexthop_cache_entry = self._update_nexthop_cache(now, vlan, None, None, ip_gw)
             last_retry_time = nexthop_cache_entry.last_retry_time
-            ip_gw_with_retry_time = (ip_gw, faucet_vip, nexthop_cache_entry)
+            ip_gw_with_retry_time = (ip_gw, nexthop_cache_entry)
             if last_retry_time is None:
                 ip_gws_never_tried.append(ip_gw_with_retry_time)
             else:
@@ -308,49 +328,35 @@ class ValveRouteManager(object):
         random.shuffle(ip_gws_never_tried)
         ip_gws_with_retry_time_sorted = list(
             sorted(ip_gws_with_retry_time, key=lambda x: x[-1].last_retry_time))
-        unresolved_nexthops = ip_gws_never_tried + ip_gws_with_retry_time_sorted
+        unresolved_nexthops = [ip_gw[0] for ip_gw in ip_gws_never_tried + ip_gws_with_retry_time_sorted]
         return unresolved_nexthops
 
-    def _is_host_fib_route(self, vlan, host_ip):
-        """Return True if IP destination is a host FIB route.
-
-        Args:
-            vlan (vlan): VLAN containing this RIB/FIB.
-            host_ip: (ipaddress.ip_address): potential host FIB route.
-        Returns:
-            True if a host FIB route (and not used as a gateway).
-        """
-        ip_dsts = vlan.ip_dsts_for_ip_gw(host_ip)
-        if (len(ip_dsts) == 1 and
-                ip_dsts[0].prefixlen == ip_dsts[0].max_prefixlen and
-                ip_dsts[0].network_address == host_ip):
-            return True
-        return False
-
     def advertise(self, vlan):
         raise NotImplementedError # pragma: no cover
 
-    def _resolve_gateway_flows(self, ip_gw, nexthop_cache_entry, vlan, faucet_vip, now):
+    def _resolve_gateway_flows(self, ip_gw, nexthop_cache_entry, vlan, now):
         resolve_flows = []
         last_retry_time = nexthop_cache_entry.last_retry_time
         nexthop_cache_entry.last_retry_time = now
         nexthop_cache_entry.resolve_retries += 1
+        faucet_vip = vlan.vip_map(ip_gw)
         resolve_flows = self.resolve_gw_on_vlan(vlan, faucet_vip, ip_gw)
         if vlan.targeted_gw_resolution:
             if last_retry_time is None and nexthop_cache_entry.port is not None:
                 resolve_flows = [self.resolve_gw_on_port(
                     vlan, nexthop_cache_entry.port, faucet_vip, ip_gw)]
-        if last_retry_time is None:
-            self.logger.info(
-                'resolving %s (%u flows) on %s' % (ip_gw, len(resolve_flows), vlan))
-        else:
-            self.logger.info(
-                'resolving %s retry %u (last attempt was %us ago; %u flows) on %s' % (
-                    ip_gw,
-                    nexthop_cache_entry.resolve_retries,
-                    now - last_retry_time,
-                    len(resolve_flows),
-                    vlan))
+        if resolve_flows:
+            if last_retry_time is None:
+                self.logger.info(
+                    'resolving %s (%u flows) on VLAN %u' % (ip_gw, len(resolve_flows), vlan.vid))
+            else:
+                self.logger.info(
+                    'resolving %s retry %u (last attempt was %us ago; %u flows) on VLAN %u' % (
+                        ip_gw,
+                        nexthop_cache_entry.resolve_retries,
+                        now - last_retry_time,
+                        len(resolve_flows),
+                        vlan.vid))
         return resolve_flows
 
     def _expire_gateway_flows(self, ip_gw, nexthop_cache_entry, vlan, now):
@@ -366,21 +372,23 @@ class ValveRouteManager(object):
             expire_flows = []
         return expire_flows
 
-    def _resolve_and_expire_gateway_flows(self, ip_gw, nexthop_cache_entry, vlan, faucet_vip, now):
+    def _resolve_expire_gateway_flows(self, ip_gw, nexthop_cache_entry, vlan, now):
         if self.nexthop_dead(nexthop_cache_entry):
             return self._expire_gateway_flows(ip_gw, nexthop_cache_entry, vlan, now)
-        return self._resolve_gateway_flows(ip_gw, nexthop_cache_entry, vlan, faucet_vip, now)
+        return self._resolve_gateway_flows(ip_gw, nexthop_cache_entry, vlan, now)
 
-    def _resolve_gateways_flows(self, resolve_handler, vlan, now, unresolved_nexthops, remaining_attempts):
+    def _resolve_gateways_flows(self, resolve_handler, vlan, now,
+                                unresolved_nexthops, remaining_attempts):
         ofmsgs = []
-        for ip_gw, faucet_vip, nexthop_cache_entry in unresolved_nexthops:
+        for ip_gw in unresolved_nexthops:
             if remaining_attempts == 0:
                 break
-            resolve_flows = resolve_handler(ip_gw, nexthop_cache_entry, vlan, faucet_vip, now)
+            nexthop_cache_entry = self._vlan_nexthop_cache_entry(vlan, ip_gw)
+            resolve_flows = resolve_handler(ip_gw, nexthop_cache_entry, vlan, now)
             if resolve_flows:
                 ofmsgs.extend(resolve_flows)
                 remaining_attempts -= 1
-        return (remaining_attempts, ofmsgs)
+        return ofmsgs
 
     def resolve_gateways(self, vlan, now):
         """Re/resolve all gateways.
@@ -391,17 +399,26 @@ class ValveRouteManager(object):
         Returns:
             list: OpenFlow messages.
         """
-        ofmsgs = []
-        remaining_attempts = self.max_hosts_per_resolve_cycle
-        route_ip_gws, host_ip_gws = self._vlan_ip_gws(vlan)
-        for ip_gws, resolve_handler in (
-                (route_ip_gws, self._resolve_gateway_flows),
-                (host_ip_gws, self._resolve_and_expire_gateway_flows)):
-            unresolved_nexthops = self._vlan_unresolved_nexthops(vlan, ip_gws, now)
-            remaining_attempts, resolve_ofmsgs = self._resolve_gateways_flows(
-                resolve_handler, vlan, now, unresolved_nexthops, remaining_attempts)
-            ofmsgs.extend(resolve_ofmsgs)
-        return ofmsgs
+        route_ip_gws, _ = self._vlan_ip_gws(vlan)
+        unresolved_nexthops = self._vlan_unresolved_nexthops(vlan, route_ip_gws, now)
+        return self._resolve_gateways_flows(
+            self._resolve_gateway_flows, vlan, now, unresolved_nexthops,
+            self.max_hosts_per_resolve_cycle)
+
+    def resolve_expire_hosts(self, vlan, now):
+        """Re/resolve all hosts.
+
+        Args:
+            vlan (vlan): VLAN containing this RIB/FIB.
+            now (float): seconds since epoch.
+        Returns:
+            list: OpenFlow messages.
+        """
+        _, host_ip_gws = self._vlan_ip_gws(vlan)
+        unresolved_nexthops = self._vlan_unresolved_nexthops(vlan, host_ip_gws, now)
+        return self._resolve_gateways_flows(
+            self._resolve_expire_gateway_flows, vlan, now, unresolved_nexthops,
+            self.max_hosts_per_resolve_cycle)
 
     def _cached_nexthop_eth_dst(self, vlan, ip_gw):
         nexthop_cache_entry = self._vlan_nexthop_cache_entry(vlan, ip_gw)
@@ -420,23 +437,27 @@ class ValveRouteManager(object):
     def _vlan_nexthop_cache_limit(self, vlan):
         raise NotImplementedError # pragma: no cover
 
-    def _proactive_resolve_neighbor(self, now, vlans, dst_ip):
+    def _proactive_resolve_neighbor(self, now, vlan, dst_ip):
         ofmsgs = []
-        if not self.proactive_learn:
-            return ofmsgs
-        for vlan in vlans:
-            faucet_vip = vlan.ip_in_vip_subnet(dst_ip)
-            if faucet_vip and not vlan.is_faucet_vip(dst_ip):
+        if self.proactive_learn:
+            router = self._router_for_vlan(vlan)
+            if router is None:
+                faucet_vip = vlan.vip_map(dst_ip)
+            else:
+                vlan, faucet_vip = router.vip_map(dst_ip)
+            if (vlan and vlan.ip_in_vip_subnet(dst_ip, faucet_vip) and
+                    faucet_vip.ip != dst_ip and self._stateful_gw(vlan, dst_ip)):
                 limit = self._vlan_nexthop_cache_limit(vlan)
                 if limit is None or len(self._vlan_nexthop_cache(vlan)) < limit:
-                    resolution_in_progress = self._is_host_fib_route(vlan, dst_ip)
+                    resolution_in_progress = vlan.is_host_fib_route(dst_ip)
                     ofmsgs.extend(self._add_host_fib_route(vlan, dst_ip, blackhole=True))
-                    nexthop_cache_entry = self._update_nexthop_cache(now, vlan, None, None, dst_ip)
+                    nexthop_cache_entry = self._update_nexthop_cache(
+                        now, vlan, None, None, dst_ip)
                     if not resolution_in_progress:
                         resolve_flows = self._resolve_gateway_flows(
-                            dst_ip, nexthop_cache_entry, vlan, faucet_vip, nexthop_cache_entry.cache_time)
+                            dst_ip, nexthop_cache_entry, vlan,
+                            nexthop_cache_entry.cache_time)
                         ofmsgs.extend(resolve_flows)
-                break
         return ofmsgs
 
     def add_route(self, vlan, ip_gw, ip_dst):
@@ -484,7 +505,8 @@ class ValveRouteManager(object):
             timeout = (
                 self.max_resolve_backoff_time * self.max_host_fib_retry_count +
                 random.randint(0, self.max_resolve_backoff_time * 2))
-            for routed_vlan in self._routed_vlans(vlan):
+            routed_vlans = self._routed_vlans(vlan)
+            for routed_vlan in routed_vlans:
                 in_match = self._route_match(routed_vlan, host_int)
                 ofmsgs.append(self.fib_table.flowmod(
                     in_match, priority=priority, hard_timeout=timeout))
@@ -533,11 +555,12 @@ class ValveRouteManager(object):
         Returns:
             list: OpenFlow messages.
         """
-        ip_pkt = self._ip_pkt(pkt_meta.pkt)
+        src_ip = pkt_meta.l3_src
         ofmsgs = []
-        if ip_pkt:
-            src_ip = ipaddress.ip_address(btos(ip_pkt.src))
-            if src_ip and pkt_meta.vlan.ip_in_vip_subnet(src_ip):
+        if (src_ip and pkt_meta.vlan.ip_in_vip_subnet(src_ip) and
+                self._stateful_gw(pkt_meta.vlan, src_ip)):
+            ip_pkt = self._ip_pkt(pkt_meta.pkt)
+            if ip_pkt:
                 ofmsgs.extend(
                     self._add_host_fib_route(pkt_meta.vlan, src_ip, blackhole=False))
                 ofmsgs.extend(self._update_nexthop(
@@ -546,13 +569,9 @@ class ValveRouteManager(object):
 
     def _del_route_flows(self, vlan, ip_dst):
         ofmsgs = []
-        for routed_vlan in self._routed_vlans(vlan):
-            if ip_dst.prefixlen == 0:
-                route_match = self.fib_table.match(
-                    vlan=routed_vlan, eth_type=self.ETH_TYPE)
-            else:
-                route_match = self.fib_table.match(
-                    vlan=routed_vlan, eth_type=self.ETH_TYPE, nw_dst=ip_dst)
+        routed_vlans = self._routed_vlans(vlan)
+        for routed_vlan in routed_vlans:
+            route_match = self._route_match(routed_vlan, ip_dst)
             ofmsgs.extend(self.fib_table.flowdel(
                 route_match, priority=self._route_priority(ip_dst), strict=True))
         return ofmsgs
@@ -575,7 +594,6 @@ class ValveRouteManager(object):
         if ip_dst in routes:
             vlan.del_route(ip_dst)
             ofmsgs.extend(self._del_route_flows(vlan, ip_dst))
-            # TODO: need to delete nexthop group if groups are in use.
         return ofmsgs
 
     def control_plane_handler(self, now, pkt_meta):
@@ -638,80 +656,77 @@ class ValveIPv4RouteManager(ValveRouteManager):
 
     def _control_plane_arp_handler(self, now, pkt_meta):
         ofmsgs = []
-        if not pkt_meta.packet_complete():
+        if not pkt_meta.eth_type == valve_of.ether.ETH_TYPE_ARP:
             return ofmsgs
-        pkt_meta.reparse_ip()
-        arp_pkt = pkt_meta.pkt.get_protocol(arp.arp)
-        if arp_pkt is None:
-            return ofmsgs
-        src_ip = ipaddress.IPv4Address(btos(arp_pkt.src_ip))
-        dst_ip = ipaddress.IPv4Address(btos(arp_pkt.dst_ip))
         vlan = pkt_meta.vlan
+        src_ip = pkt_meta.l3_src
+        dst_ip = pkt_meta.l3_dst
         if vlan.from_connected_to_vip(src_ip, dst_ip):
+            arp_pkt = pkt_meta.pkt.get_protocol(arp.arp)
+            if arp_pkt is None:
+                return ofmsgs
             opcode = arp_pkt.opcode
             port = pkt_meta.port
             eth_src = pkt_meta.eth_src
             if opcode == arp.ARP_REQUEST:
-                ofmsgs.extend(
-                    self._add_host_fib_route(vlan, src_ip, blackhole=False))
-                ofmsgs.extend(self._update_nexthop(
-                    now, vlan, port, eth_src, src_ip))
-                ofmsgs.append(
-                    vlan.pkt_out_port(
-                        valve_packet.arp_reply, port,
-                        vlan.faucet_mac, eth_src, dst_ip, src_ip))
-                self.logger.info(
-                    'Responded to ARP request for %s from %s (%s) on VLAN %u' % (
-                        dst_ip, src_ip, eth_src, vlan.vid))
-            elif (opcode == arp.ARP_REPLY and
-                  pkt_meta.eth_dst == vlan.faucet_mac):
-                ofmsgs.extend(
-                    self._update_nexthop(now, vlan, port, eth_src, src_ip))
-                self.logger.info(
-                    'ARP response %s (%s) on VLAN %u' % (
-                        src_ip, eth_src, vlan.vid))
+                if pkt_meta.eth_dst == valve_of.mac.BROADCAST_STR:
+                    if self._stateful_gw(vlan, src_ip):
+                        ofmsgs.extend(
+                            self._add_host_fib_route(vlan, src_ip, blackhole=False))
+                        ofmsgs.extend(self._update_nexthop(
+                            now, vlan, port, eth_src, src_ip))
+                    ofmsgs.append(
+                        vlan.pkt_out_port(
+                            valve_packet.arp_reply, port,
+                            vlan.faucet_mac, eth_src, dst_ip, src_ip))
+                    self.logger.info(
+                        'Responded to ARP request for %s from %s' % (
+                            dst_ip, pkt_meta.log()))
+            elif opcode == arp.ARP_REPLY:
+                if pkt_meta.eth_dst == vlan.faucet_mac:
+                    ofmsgs.extend(
+                        self._update_nexthop(now, vlan, port, eth_src, src_ip))
+                    self.logger.info(
+                        'Received ARP response %s from %s' % (
+                            src_ip, pkt_meta.log()))
         return ofmsgs
 
     def _control_plane_icmp_handler(self, pkt_meta, ipv4_pkt):
         ofmsgs = []
-        if not pkt_meta.packet_complete():
+        if ipv4_pkt.proto != valve_of.inet.IPPROTO_ICMP:
             return ofmsgs
-
-        src_ip = ipaddress.IPv4Address(btos(ipv4_pkt.src))
-        dst_ip = ipaddress.IPv4Address(btos(ipv4_pkt.dst))
         vlan = pkt_meta.vlan
+        if pkt_meta.eth_dst != vlan.faucet_mac:
+            return ofmsgs
+        src_ip = pkt_meta.l3_src
+        dst_ip = pkt_meta.l3_dst
         if vlan.from_connected_to_vip(src_ip, dst_ip):
-            if pkt_meta.eth_dst != vlan.faucet_mac:
-                return ofmsgs
-            if ipv4_pkt.proto != valve_of.inet.IPPROTO_ICMP:
-                return ofmsgs
             pkt_meta.reparse_all()
             icmp_pkt = pkt_meta.pkt.get_protocol(icmp.icmp)
             if icmp_pkt is None:
                 return ofmsgs
             if icmp_pkt.type == icmp.ICMP_ECHO_REQUEST:
-                port = pkt_meta.port
                 ofmsgs.append(
                     vlan.pkt_out_port(
-                        valve_packet.echo_reply, port,
+                        valve_packet.echo_reply, pkt_meta.port,
                         vlan.faucet_mac, pkt_meta.eth_src,
                         dst_ip, src_ip, icmp_pkt.data))
         return ofmsgs
 
     def control_plane_handler(self, now, pkt_meta):
-        ipv4_pkt = pkt_meta.pkt.get_protocol(ipv4.ipv4)
-        if ipv4_pkt is None:
-            return self._control_plane_arp_handler(now, pkt_meta)
-        else:
+        if pkt_meta.packet_complete():
+            arp_replies = self._control_plane_arp_handler(now, pkt_meta)
+            if arp_replies:
+                return arp_replies
+            ipv4_pkt = pkt_meta.pkt.get_protocol(ipv4.ipv4)
+            if ipv4_pkt is None:
+                return []
             icmp_replies = self._control_plane_icmp_handler(
                 pkt_meta, ipv4_pkt)
             if icmp_replies:
                 return icmp_replies
-            dst_ip = ipaddress.IPv4Address(btos(ipv4_pkt.dst))
-            vlan = pkt_meta.vlan
-            return self._proactive_resolve_neighbor(
-                now, self._routed_vlans(vlan), dst_ip)
-        return []
+        return self._proactive_resolve_neighbor(
+            now, pkt_meta.vlan, pkt_meta.l3_dst)
 
 
 class ValveIPv6RouteManager(ValveRouteManager):
@@ -774,7 +789,7 @@ class ValveIPv6RouteManager(ValveRouteManager):
                 icmpv6_type=icmpv6.ND_NEIGHBOR_ADVERT),
             priority=priority,
             max_len=self.MAX_LEN))
-        if faucet_vip.ip in valve_packet.IPV6_LINK_LOCAL:
+        if faucet_vip.ip.is_link_local:
             ofmsgs.append(self.eth_src_table.flowmod(
                 self.eth_src_table.match(
                     eth_type=self.ETH_TYPE,
@@ -795,61 +810,63 @@ class ValveIPv6RouteManager(ValveRouteManager):
         ofmsgs = super(ValveIPv6RouteManager, self)._add_faucet_fib_to_vip(
             vlan, priority, faucet_vip, faucet_vip_host)
         faucet_vip_broadcast = ipaddress.IPv6Interface(faucet_vip.network.broadcast_address)
+        if self._global_routing():
+            vlan = self.global_vlan
         ofmsgs.append(self.fib_table.flowmod(
-            self.fib_table.match(eth_type=self.ETH_TYPE, vlan=vlan, nw_dst=faucet_vip_broadcast),
+            self._route_match(vlan, faucet_vip_broadcast),
             priority=priority,
             inst=[valve_of.goto_table(self.vip_table)]))
         return ofmsgs
 
     def _nd_solicit_handler(self, now, pkt_meta, _ipv6_pkt, icmpv6_pkt, src_ip, _dst_ip):
         ofmsgs = []
-        solicited_ip = btos(icmpv6_pkt.data.dst)
-        vlan = pkt_meta.vlan
-        if vlan.is_faucet_vip(ipaddress.ip_address(solicited_ip)):
-            ofmsgs.extend(
-                self._add_host_fib_route(vlan, src_ip, blackhole=False))
-            ofmsgs.extend(self._update_nexthop(
-                now, vlan, pkt_meta.port, pkt_meta.eth_src, src_ip))
-            ofmsgs.append(
-                vlan.pkt_out_port(
-                    valve_packet.nd_advert, pkt_meta.port,
-                    vlan.faucet_mac, pkt_meta.eth_src,
-                    solicited_ip, src_ip))
-            self.logger.info(
-                'Responded to ND solicit for %s to %s (%s) on VLAN %u' % (
-                    solicited_ip, src_ip, pkt_meta.eth_src, vlan.vid))
+        if isinstance(icmpv6_pkt.data, icmpv6.nd_neighbor):
+            solicited_ip = ipaddress.ip_address(btos(icmpv6_pkt.data.dst))
+            vlan = pkt_meta.vlan
+            if vlan.is_faucet_vip(solicited_ip):
+                if self._stateful_gw(vlan, src_ip):
+                    ofmsgs.extend(
+                        self._add_host_fib_route(vlan, src_ip, blackhole=False))
+                    ofmsgs.extend(self._update_nexthop(
+                        now, vlan, pkt_meta.port, pkt_meta.eth_src, src_ip))
+                ofmsgs.append(
+                    vlan.pkt_out_port(
+                        valve_packet.nd_advert, pkt_meta.port,
+                        vlan.faucet_mac, pkt_meta.eth_src,
+                        solicited_ip, src_ip))
+                self.logger.info(
+                    'Responded to ND solicit for %s from %s' % (
+                        solicited_ip, pkt_meta.log()))
         return ofmsgs
 
     def _nd_advert_handler(self, now, pkt_meta, _ipv6_pkt, icmpv6_pkt, _src_ip, _dst_ip):
         ofmsgs = []
-        target_ip = ipaddress.ip_address(btos(icmpv6_pkt.data.dst))
-        vlan = pkt_meta.vlan
-        if vlan.ip_in_vip_subnet(target_ip):
-            ofmsgs.extend(self._update_nexthop(
-                now, vlan, pkt_meta.port, pkt_meta.eth_src, target_ip))
-            self.logger.info(
-                'ND advert %s (%s) on VLAN %u' % (
-                    target_ip, pkt_meta.eth_src, vlan.vid))
+        if isinstance(icmpv6_pkt.data, icmpv6.nd_neighbor):
+            target_ip = ipaddress.ip_address(btos(icmpv6_pkt.data.dst))
+            vlan = pkt_meta.vlan
+            if vlan.ip_in_vip_subnet(target_ip):
+                if self._stateful_gw(vlan, target_ip):
+                    ofmsgs.extend(self._update_nexthop(
+                        now, vlan, pkt_meta.port, pkt_meta.eth_src, target_ip))
+                self.logger.info(
+                    'Received ND advert for %s from %s' % (
+                        target_ip, pkt_meta.log()))
         return ofmsgs
 
-    def _router_solicit_handler(self, now, pkt_meta, _ipv6_pkt, _icmpv6_pkt, src_ip, _dst_ip):
+    def _router_solicit_handler(self, _now, pkt_meta, _ipv6_pkt, _icmpv6_pkt, src_ip, _dst_ip):
         ofmsgs = []
         vlan = pkt_meta.vlan
-        link_local_vips, other_vips = self._link_and_other_vips(vlan)
+        link_local_vips, other_vips = vlan.link_and_other_vips(self.IPV)
         for vip in link_local_vips:
             if src_ip in vip.network:
-                ofmsgs.extend(
-                    self._add_host_fib_route(vlan, src_ip, blackhole=False))
-                ofmsgs.extend(self._update_nexthop(
-                    now, vlan, pkt_meta.port, pkt_meta.eth_src, src_ip))
                 ofmsgs.append(
                     vlan.pkt_out_port(
                         valve_packet.router_advert, pkt_meta.port,
                         vlan.faucet_mac, pkt_meta.eth_src,
                         vip.ip, src_ip, other_vips))
                 self.logger.info(
-                    'Responded to RS solicit from %s (%s) to VIP %s on VLAN %u' % (
-                        src_ip, pkt_meta. eth_src, vip, vlan.vid))
+                    'Responded to RS solicit from %s (%s)' % (
+                        src_ip, pkt_meta.log()))
                 break
         return ofmsgs
 
@@ -857,7 +874,8 @@ class ValveIPv6RouteManager(ValveRouteManager):
         ofmsgs = []
         vlan = pkt_meta.vlan
         if (vlan.from_connected_to_vip(src_ip, dst_ip) and
-                pkt_meta.eth_dst == vlan.faucet_mac):
+                pkt_meta.eth_dst == vlan.faucet_mac and
+                isinstance(icmpv6_pkt.data, icmpv6.echo)):
             ofmsgs.append(
                 vlan.pkt_out_port(
                     valve_packet.icmpv6_echo_reply, pkt_meta.port,
@@ -876,18 +894,16 @@ class ValveIPv6RouteManager(ValveRouteManager):
 
     def _control_plane_icmpv6_handler(self, now, pkt_meta, ipv6_pkt):
         ofmsgs = []
-        if not pkt_meta.packet_complete():
-            return ofmsgs
         # Must be ICMPv6 and have no extended headers.
         if ipv6_pkt.nxt != valve_of.inet.IPPROTO_ICMPV6:
             return ofmsgs
         if ipv6_pkt.ext_hdrs:
             return ofmsgs
-        dst_ip = ipaddress.IPv6Address(btos(ipv6_pkt.dst))
+        dst_ip = pkt_meta.l3_dst
         # Explicitly ignore messages to all notes.
         if dst_ip == valve_packet.IPV6_ALL_NODES:
             return ofmsgs
-        src_ip = ipaddress.IPv6Address(btos(ipv6_pkt.src))
+        src_ip = pkt_meta.l3_src
         vlan = pkt_meta.vlan
         if not vlan.ip_in_vip_subnet(src_ip):
             return ofmsgs
@@ -905,31 +921,20 @@ class ValveIPv6RouteManager(ValveRouteManager):
         return ofmsgs
 
     def control_plane_handler(self, now, pkt_meta):
-        pkt = pkt_meta.pkt
-        ipv6_pkt = pkt.get_protocol(ipv6.ipv6)
-        if ipv6_pkt is not None:
-            icmp_replies = self._control_plane_icmpv6_handler(
-                now, pkt_meta, ipv6_pkt)
-            if icmp_replies:
-                return icmp_replies
-            dst_ip = ipaddress.IPv6Address(btos(ipv6_pkt.dst))
-            return self._proactive_resolve_neighbor(
-                now, self._routed_vlans(pkt_meta.vlan), dst_ip)
+        if pkt_meta.packet_complete():
+            ipv6_pkt = pkt_meta.pkt.get_protocol(ipv6.ipv6)
+            if ipv6_pkt is not None:
+                icmp_replies = self._control_plane_icmpv6_handler(
+                    now, pkt_meta, ipv6_pkt)
+                if icmp_replies:
+                    return icmp_replies
+                return self._proactive_resolve_neighbor(
+                    now, pkt_meta.vlan, pkt_meta.l3_dst)
         return []
 
-    def _link_and_other_vips(self, vlan):
-        link_local_vips = []
-        other_vips = []
-        for faucet_vip in vlan.faucet_vips_by_ipv(self.IPV):
-            if faucet_vip.ip in valve_packet.IPV6_LINK_LOCAL:
-                link_local_vips.append(faucet_vip)
-            else:
-                other_vips.append(faucet_vip)
-        return link_local_vips, other_vips
-
     def advertise(self, vlan):
         ofmsgs = []
-        link_local_vips, other_vips = self._link_and_other_vips(vlan)
+        link_local_vips, other_vips = vlan.link_and_other_vips(self.IPV)
         for link_local_vip in link_local_vips:
             # https://tools.ietf.org/html/rfc4861#section-6.1.2
             ofmsgs.extend(vlan.flood_pkt(
