@@ -22,20 +22,27 @@ import struct
 from faucet import valve_of
 
 
-class ValveTable(object):
+class ValveTable: # pylint: disable=too-many-arguments,too-many-instance-attributes
     """Wrapper for an OpenFlow table."""
 
-    def __init__(self, table_id, name, restricted_match_types,
+    def __init__(self, table_id, name, table_config,
                  flow_cookie, notify_flow_removed=False):
         self.table_id = table_id
         self.name = name
-        self.restricted_match_types = None
-        if restricted_match_types:
-            self.restricted_match_types = set(restricted_match_types)
+        self.table_config = table_config
+        self.set_fields = self.table_config.set_fields
+        self.exact_match = self.table_config.exact_match
+        self.match_types = None
+        if self.table_config.match_types:
+            self.match_types = {}
+            for field, mask in self.table_config.match_types:
+                self.match_types[field] = mask
         self.flow_cookie = flow_cookie
         self.notify_flow_removed = notify_flow_removed
 
-    def match(self, in_port=None, vlan=None,
+    # TODO: verify set_fields
+    # TODO: verify actions
+    def match(self, in_port=None, vlan=None, # pylint: disable=too-many-arguments
               eth_type=None, eth_src=None,
               eth_dst=None, eth_dst_mask=None,
               icmpv6_type=None,
@@ -45,21 +52,37 @@ class ValveTable(object):
             in_port, vlan, eth_type, eth_src,
             eth_dst, eth_dst_mask, icmpv6_type,
             nw_proto, nw_dst)
-        match = valve_of.match(match_dict)
-        if self.restricted_match_types is not None:
-            for match_type in match_dict:
-                assert match_type in self.restricted_match_types, '%s match in table %s' % (
-                    match_type, self.name)
-        return match
+        return valve_of.match(match_dict)
 
-    def flowmod(self, match=None, priority=None,
+    def _verify_flowmod(self, flowmod):
+        if valve_of.is_flowdel(flowmod):
+            return
+        if flowmod.priority == 0:
+            assert not flowmod.match.items(), (
+                'default flow cannot have matches')
+        elif self.match_types:
+            match_fields = list(flowmod.match.items())
+            for match_type, match_field in match_fields:
+                assert match_type in self.match_types, (
+                    '%s match in table %s' % (match_type, self.name))
+                config_mask = self.match_types[match_type]
+                flow_mask = isinstance(match_field, tuple)
+                assert config_mask or (not config_mask and not flow_mask), (
+                    '%s configured mask %s but flow mask %s in table %s (%s)' % (
+                        match_type, config_mask, flow_mask, self.name, flowmod))
+            if self.exact_match:
+                assert len(self.match_types) == len(match_fields), (
+                    'exact match table matches %s do not match flow matches %s (%s)' % (
+                        self.match_types, match_fields, flowmod))
+
+    def flowmod(self, match=None, priority=None, # pylint: disable=too-many-arguments
                 inst=None, command=valve_of.ofp.OFPFC_ADD, out_port=0,
                 out_group=0, hard_timeout=0, idle_timeout=0, cookie=None):
         """Helper function to construct a flow mod message with cookie."""
-        if match is None:
-            match = self.match()
         if priority is None:
             priority = 0 # self.dp.lowest_priority
+        if not match:
+            match = self.match()
         if inst is None:
             inst = []
         if cookie is None:
@@ -67,7 +90,7 @@ class ValveTable(object):
         flags = 0
         if self.notify_flow_removed:
             flags = valve_of.ofp.OFPFF_SEND_FLOW_REM
-        return valve_of.flowmod(
+        flowmod = valve_of.flowmod(
             cookie,
             command,
             self.table_id,
@@ -79,6 +102,8 @@ class ValveTable(object):
             hard_timeout,
             idle_timeout,
             flags)
+        self._verify_flowmod(flowmod)
+        return flowmod
 
     def flowdel(self, match=None, priority=None, out_port=valve_of.ofp.OFPP_ANY, strict=False):
         """Delete matching flows from a table."""
@@ -112,7 +137,7 @@ class ValveTable(object):
                 [valve_of.output_controller(max_len)])] + inst)
 
 
-class ValveGroupEntry(object):
+class ValveGroupEntry:
     """Abstraction for a single OpenFlow group entry."""
 
     def __init__(self, table, group_id, buckets):
@@ -121,6 +146,7 @@ class ValveGroupEntry(object):
         self.update_buckets(buckets)
 
     def update_buckets(self, buckets):
+        """Update entry with new buckets."""
         self.buckets = tuple(buckets)
 
     def add(self):
@@ -145,10 +171,15 @@ class ValveGroupEntry(object):
         return valve_of.groupdel(group_id=self.group_id)
 
 
-class ValveGroupTable(object):
+class ValveGroupTable:
     """Wrap access to group table."""
 
-    entries = {} # type: dict
+    entries = None # type: dict
+
+
+    def __init__(self):
+        """Constructs a new object"""
+        self.entries = {}
 
     @staticmethod
     def group_id_from_str(key_str):
@@ -158,6 +189,7 @@ class ValveGroupTable(object):
         return struct.unpack('<L', digest[:4])[0]
 
     def get_entry(self, group_id, buckets):
+        """Update entry with group_id with buckets, and return the entry."""
         if group_id in self.entries:
             self.entries[group_id].update_buckets(buckets)
         else:
