@@ -24,7 +24,7 @@ from faucet.valve import valve_factory, SUPPORTED_HARDWARE
 from faucet.valve_util import dpid_log, stat_config_files
 
 
-class ConfigWatcher(object):
+class ConfigWatcher:
     """Watch config for file or content changes."""
 
     config_file = None
@@ -55,10 +55,10 @@ class ConfigWatcher(object):
         self.config_hashes = new_config_hashes
 
 
-class ValvesManager(object):
+class ValvesManager:
     """Manage a collection of Valves."""
 
-    valves = {} # type: dict
+    valves = None # type: dict
 
     def __init__(self, logname, logger, metrics, notifier, bgp,
                  dot1x, send_flows_to_dp_by_id):
@@ -79,6 +79,7 @@ class ValvesManager(object):
         self.bgp = bgp
         self.dot1x = dot1x
         self.send_flows_to_dp_by_id = send_flows_to_dp_by_id
+        self.valves = {}
         self.config_watcher = ConfigWatcher()
 
     def parse_configs(self, new_config_file):
@@ -149,13 +150,14 @@ class ValvesManager(object):
     def valve_flow_services(self, now, valve_service):
         """Call a method on all Valves and send any resulting flows."""
         for valve in list(self.valves.values()):
-            ofmsgs = getattr(valve, valve_service)(now)
+            other_valves = self._other_running_valves(valve)
+            ofmsgs = getattr(valve, valve_service)(now, other_valves)
             if ofmsgs:
                 self.send_flows_to_dp_by_id(valve, ofmsgs)
 
     def _other_running_valves(self, valve):
         return [other_valve for other_valve in list(self.valves.values())
-                if valve != other_valve and other_valve.dp.running]
+                if valve != other_valve and other_valve.dp.dyn_running]
 
     def valve_packet_in(self, now, valve, msg):
         """Time a call to Valve packet in handler."""
@@ -167,14 +169,8 @@ class ValvesManager(object):
         self.metrics.of_packet_ins.labels( # pylint: disable=no-member
             **valve.base_prom_labels).inc()
         with self.metrics.faucet_packet_in_secs.labels( # pylint: disable=no-member
-            **valve.base_prom_labels).time():
+                **valve.base_prom_labels).time():
             ofmsgs = valve.rcv_packet(now, self._other_running_valves(valve), pkt_meta)
         if ofmsgs:
             self.send_flows_to_dp_by_id(valve, ofmsgs)
             valve.update_metrics(now, pkt_meta.port, rate_limited=True)
-
-    def stack_topo_change(self, _now, valve):
-        """Update stack topo of all other Valves affected by the event on this Valve."""
-        for other_valve in self._other_running_valves(valve):
-            other_valve.flood_manager.update_stack_topo(valve.dp.running, valve)
-            # TODO: rebuild flood rules
