@@ -28,16 +28,31 @@ from faucet.valve_util import btos
 from faucet.valve_packet import FAUCET_MAC
 
 
-class NullVLAN(object):
+class NullVLAN:
     """Placeholder null VLAN."""
 
     name = 'Null VLAN'
     vid = valve_of.ofp.OFPVID_NONE
 
 
-class HostCacheEntry(object):
+# TODO: not well supported by any hardware, so not used.
+class AnyVLAN:
+    """Placeholder any tagged VLAN."""
+
+    name = 'Any VLAN'
+    vid = valve_of.ofp.OFPVID_PRESENT
+
+
+class HostCacheEntry:
     """Association of a host with a port."""
 
+    __slots__ = [
+        'cache_time',
+        'eth_src',
+        'eth_src_int',
+        'port',
+    ]
+
     def __init__(self, eth_src, port, cache_time):
         self.eth_src = eth_src
         self.port = port
@@ -66,44 +81,7 @@ class VLAN(Conf):
 # Note: while vlans are configured once for each datapath, there will be a
 # separate vlan object created for each datapath that the vlan appears on
 
-    name = None
-    dp_id = None
-    tagged = None
-    untagged = None
-    vid = None
-    faucet_vips = None
-    faucet_mac = None
-    bgp_as = None
-    bgp_connect_mode = None
-    bgp_local_address = None
-    bgp_server_addresses = [] # type: list
-    bgp_port = None
-    bgp_routerid = None
-    bgp_neighbor_addresses = [] # type: list
-    bgp_neighbour_addresses = [] # type: list
-    bgp_neighbor_as = None
-    bgp_neighbour_as = None
-    routes = None
-    max_hosts = None
-    unicast_flood = None
-    acl_in = None
-    acls_in = None
-    targeted_gw_resolution = None
-    proactive_arp_limit = None
-    proactive_nd_limit = None
-    # Define dynamic variables with prefix dyn_ to distinguish from variables set
-    # configuration
-    dyn_host_cache = None
-    dyn_host_cache_by_port = None
-    dyn_faucet_vips_by_ipv = None
-    dyn_bgp_server_addresses_by_ipv = None
-    dyn_bgp_neighbor_addresses_by_ipv = None
-    dyn_routes_by_ipv = None
-    dyn_gws_by_ipv = None
-    dyn_neigh_cache_by_ipv = None
-    dyn_learn_ban_count = 0
-    dyn_last_time_hosts_expired = None
-    dyn_oldest_host_time = None
+    mutable_attrs = frozenset(['tagged', 'untagged'])
 
     defaults = {
         'name': None,
@@ -166,15 +144,50 @@ class VLAN(Conf):
     }
 
     def __init__(self, _id, dp_id, conf=None):
+        self.acl_in = None
+        self.acls_in = None
+        self.bgp_as = None
+        self.bgp_connect_mode = None
+        self.bgp_local_address = None
+        self.bgp_neighbor_addresses = None
+        self.bgp_neighbor_as = None
+        self.bgp_neighbour_addresses = None
+        self.bgp_neighbour_as = None
+        self.bgp_port = None
+        self.bgp_routerid = None
+        self.bgp_server_addresses = None
+        self.description = None
+        self.dp_id = None
+        self.faucet_mac = None
+        self.faucet_vips = None
+        self.max_hosts = None
+        self.minimum_ip_size_check = None
+        self.name = None
+        self.proactive_arp_limit = None
+        self.proactive_nd_limit = None
+        self.routes = None
+        self.tagged = None
+        self.targeted_gw_resolution = None
+        self.unicast_flood = None
+        self.untagged = None
+        self.vid = None
+
+        self.acls = {}
         self.tagged = []
         self.untagged = []
-        self.dyn_faucet_vips_by_ipv = collections.defaultdict(list)
-        self.dyn_bgp_server_addresses_by_ipv = collections.defaultdict(list)
-        self.dyn_bgp_neighbor_addresses_by_ipv = collections.defaultdict(list)
+        self.bgp_server_addresses = []
+        self.bgp_neighbour_addresses = []
+        self.bgp_neighbor_addresses = []
+
+        self.dyn_host_cache = None
+        self.dyn_host_cache_by_port = None
+        self.dyn_last_time_hosts_expired = None
+        self.dyn_learn_ban_count = 0
+        self.dyn_neigh_cache_by_ipv = None
+        self.dyn_oldest_host_time = None
+
         self.dyn_routes_by_ipv = collections.defaultdict(dict)
         self.dyn_gws_by_ipv = collections.defaultdict(dict)
-        self.dyn_ipvs = []
-        self.dyn_bgp_ipvs = []
         self.reset_caches()
         super(VLAN, self).__init__(_id, dp_id, conf)
 
@@ -215,30 +228,21 @@ class VLAN(Conf):
                 self.proactive_nd_limit = 2 * self.max_hosts
 
         if self.faucet_vips:
-            self.faucet_vips = [
-                self._check_ip_str(ip_str, ip_method=ipaddress.ip_interface) for ip_str in self.faucet_vips]
-            for faucet_vip in self.faucet_vips:
-                self.dyn_faucet_vips_by_ipv[faucet_vip.version].append(faucet_vip)
-            self.dyn_ipvs = list(self.dyn_faucet_vips_by_ipv.keys())
+            self.faucet_vips = frozenset([
+                self._check_ip_str(ip_str, ip_method=ipaddress.ip_interface) for ip_str in self.faucet_vips])
 
         if self.bgp_neighbor_addresses or self.bgp_neighbour_addresses:
-            neigh_addresses = set(self.bgp_neighbor_addresses + self.bgp_neighbour_addresses)
-            self.bgp_neighbor_addresses = [
-                self._check_ip_str(ip_str) for ip_str in neigh_addresses]
-            for bgp_neighbor_address in self.bgp_neighbor_addresses:
-                self.dyn_bgp_neighbor_addresses_by_ipv[bgp_neighbor_address.version].append(
-                    bgp_neighbor_address)
+            neigh_addresses = frozenset(self.bgp_neighbor_addresses + self.bgp_neighbour_addresses)
+            self.bgp_neighbor_addresses = frozenset([
+                self._check_ip_str(ip_str) for ip_str in neigh_addresses])
 
         if self.bgp_server_addresses:
-            self.bgp_server_addresses = [
-                self._check_ip_str(ip_str) for ip_str in self.bgp_server_addresses]
-            for bgp_server_address in self.bgp_server_addresses:
-                self.dyn_bgp_server_addresses_by_ipv[bgp_server_address.version].append(
-                    bgp_server_address)
+            self.bgp_server_addresses = frozenset([
+                self._check_ip_str(ip_str) for ip_str in self.bgp_server_addresses])
+            for ipv in self.bgp_ipvs():
                 test_config_condition(
-                    len(self.dyn_bgp_server_addresses_by_ipv[bgp_server_address.version]) != 1,
+                    len(self.bgp_server_addresses_by_ipv(ipv)) != 1,
                     'Only one BGP server address per IP version supported')
-            self.dyn_bgp_ipvs = list(self.dyn_bgp_server_addresses_by_ipv.keys())
 
         if self.bgp_as:
             test_config_condition(not isinstance(self.bgp_port, int), (
@@ -253,16 +257,19 @@ class VLAN(Conf):
                 'Must be as many BGP neighbor addresses as BGP server addresses'))
 
         if self.routes:
+            test_config_condition(not isinstance(self.routes, list), 'invalid VLAN routes format')
             try:
                 self.routes = [route['route'] for route in self.routes]
             except TypeError:
                 raise InvalidConfigError('%s is not a valid routes value' % self.routes)
+            except KeyError:
+                pass
             for route in self.routes:
-                try:
-                    ip_gw = self._check_ip_str(route['ip_gw'])
-                    ip_dst = self._check_ip_str(route['ip_dst'], ip_method=ipaddress.ip_network)
-                except KeyError as err:
-                    raise InvalidConfigError('missing route config %s' % err)
+                test_config_condition(not isinstance(route, dict), 'invalid VLAN route format')
+                test_config_condition('ip_gw' not in route, 'missing ip_gw in VLAN route')
+                test_config_condition('ip_dst' not in route, 'missing ip_dst in VLAN route')
+                ip_gw = self._check_ip_str(route['ip_gw'])
+                ip_dst = self._check_ip_str(route['ip_dst'], ip_method=ipaddress.ip_network)
                 test_config_condition(
                     ip_gw.version != ip_dst.version,
                     'ip_gw version does not match the ip_dst version')
@@ -282,8 +289,8 @@ class VLAN(Conf):
         self.dyn_neigh_cache_by_ipv = collections.defaultdict(dict)
 
     def reset_ports(self, ports):
-        self.tagged = [port for port in ports if self in port.tagged_vlans]
-        self.untagged = [port for port in ports if self == port.native_vlan]
+        self.tagged = tuple([port for port in ports if self in port.tagged_vlans])
+        self.untagged = tuple([port for port in ports if self == port.native_vlan])
 
     def add_cache_host(self, eth_src, port, cache_time):
         existing_entry = self.cached_host(eth_src)
@@ -349,25 +356,39 @@ class VLAN(Conf):
                 self.expire_cache_host(entry.eth_src)
         return expired_hosts
 
+    @staticmethod
+    def _ipvs(ipas):
+        return frozenset([ipa.version for ipa in ipas])
+
+    @staticmethod
+    def _by_ipv(ipas, ipv):
+        return frozenset([ipa for ipa in ipas if ipa.version == ipv])
+
+    def faucet_vips_by_ipv(self, ipv):
+        """Return VIPs with specified IP version on this VLAN."""
+        return self._by_ipv(self.faucet_vips, ipv)
+
+    def link_and_other_vips(self, ipv):
+        vips = self.faucet_vips_by_ipv(ipv)
+        link_local_vips = frozenset([vip for vip in vips if vip.is_link_local])
+        other_vips = vips - link_local_vips
+        return (link_local_vips, other_vips)
+
+    def bgp_neighbor_addresses_by_ipv(self, ipv):
+        """Return BGP neighbor addresses with specified IP version on this VLAN."""
+        return self._by_ipv(self.bgp_neighbor_addresses, ipv)
+
+    def bgp_server_addresses_by_ipv(self, ipv):
+        """Return BGP server addresses with specified IP version on this VLAN."""
+        return self._by_ipv(self.bgp_server_addresses, ipv)
+
     def ipvs(self):
-        """Return list of IP versions configured on this VLAN."""
-        return self.dyn_ipvs
+        """Return IP versions configured on this VLAN."""
+        return self._ipvs(self.faucet_vips)
 
     def bgp_ipvs(self):
         """Return list of IP versions for BGP configured on this VLAN."""
-        return self.dyn_bgp_ipvs
-
-    def faucet_vips_by_ipv(self, ipv):
-        """Return list of VIPs with specified IP version on this VLAN."""
-        return self.dyn_faucet_vips_by_ipv[ipv]
-
-    def bgp_neighbor_addresses_by_ipv(self, ipv):
-        """Return list of BGP neighbor addresses with specified IP version on this VLAN."""
-        return self.dyn_bgp_neighbor_addresses_by_ipv[ipv]
-
-    def bgp_server_addresses_by_ipv(self, ipv):
-        """Return list of BGP server addresses with specified IP version on this VLAN."""
-        return self.dyn_bgp_server_addresses_by_ipv[ipv]
+        return self._ipvs(self.bgp_server_addresses)
 
     def routes_by_ipv(self, ipv):
         """Return route table for specified IP version on this VLAN."""
@@ -377,6 +398,21 @@ class VLAN(Conf):
         """Return route table count for specified IP version on this VLAN."""
         return len(self.dyn_routes_by_ipv[ipv])
 
+    def is_host_fib_route(self, host_ip):
+        """Return True if IP destination is a host FIB route.
+
+        Args:
+            host_ip: (ipaddress.ip_address): potential host FIB route.
+        Returns:
+            True if a host FIB route (and not used as a gateway).
+        """
+        ip_dsts = self.ip_dsts_for_ip_gw(host_ip)
+        if (len(ip_dsts) == 1 and
+                ip_dsts[0].prefixlen == ip_dsts[0].max_prefixlen and
+                ip_dsts[0].network_address == host_ip):
+            return True
+        return False
+
     def add_route(self, ip_dst, ip_gw):
         """Add an IP route."""
         self.dyn_routes_by_ipv[ip_gw.version][ip_dst] = ip_gw
@@ -399,8 +435,8 @@ class VLAN(Conf):
         return []
 
     def all_ip_gws(self, ipv):
-        """Return list of all IP gateways for specified IP version."""
-        return list(self.dyn_gws_by_ipv[ipv].keys())
+        """Return all IP gateways for specified IP version."""
+        return frozenset(self.dyn_gws_by_ipv[ipv].keys())
 
     def neigh_cache_by_ipv(self, ipv):
         """Return neighbor cache for specified IP version on this VLAN."""
@@ -415,7 +451,7 @@ class VLAN(Conf):
         return len(self.dyn_host_cache)
 
     def __str__(self):
-        port_list = [str(x) for x in self.get_ports()]
+        port_list = tuple([str(x) for x in self.get_ports()])
         ports = ','.join(port_list)
         return 'VLAN %s vid:%s ports:%s' % (self.name, self.vid, ports)
 
@@ -423,28 +459,29 @@ class VLAN(Conf):
         return self.__str__()
 
     def get_ports(self):
-        """Return list of all ports on this VLAN."""
-        return list(self.tagged) + list(self.untagged)
+        """Return all ports on this VLAN."""
+        return self.tagged + self.untagged
 
     def hairpin_ports(self):
         """Return all ports with hairpin enabled."""
-        return [port for port in self.get_ports() if port.hairpin]
+        return tuple([port for port in self.get_ports() if port.hairpin])
 
     def mirrored_ports(self):
         """Return list of ports that are mirrored on this VLAN."""
-        return [port for port in self.get_ports() if port.mirror]
+        return tuple([port for port in self.get_ports() if port.mirror])
 
     def lags(self):
         """Return dict of LAGs mapped to member ports."""
-        lacp_ports = [port for port in self.get_ports() if port.lacp]
+        lacp_ports = tuple([port for port in self.get_ports() if port.lacp])
         lags = collections.defaultdict(list)
         for port in lacp_ports:
             lags[port.lacp].append(port)
         return lags
 
-    def flood_ports(self, configured_ports, exclude_unicast):
+    @staticmethod
+    def flood_ports(configured_ports, exclude_unicast):
         if exclude_unicast:
-            return [port for port in configured_ports if port.unicast_flood]
+            return tuple([port for port in configured_ports if port.unicast_flood])
         return configured_ports
 
     def tagged_flood_ports(self, exclude_unicast):
@@ -491,21 +528,27 @@ class VLAN(Conf):
         """Return True if port number is an untagged port on this VLAN."""
         return port in self.untagged
 
-    def is_faucet_vip(self, ipa):
-        """Return True if IP is a VIP on this VLAN."""
-        for faucet_vip in self.faucet_vips_by_ipv(ipa.version):
-            if ipa == faucet_vip.ip:
-                return True
-        return False
-
-    def ip_in_vip_subnet(self, ipa):
-        """Return faucet_vip if IP in same IP network as a VIP on this VLAN."""
-        for faucet_vip in self.faucet_vips_by_ipv(ipa.version):
+    def vip_map(self, ipa):
+        for faucet_vip in self.faucet_vips:
             if ipa in faucet_vip.network:
-                if ipa not in (
-                        faucet_vip.network.network_address,
-                        faucet_vip.network.broadcast_address):
-                    return faucet_vip
+                return faucet_vip
+        return None
+
+    def is_faucet_vip(self, ipa, faucet_vip=None):
+        """Return True if IP is a VIP on this VLAN."""
+        if faucet_vip is None:
+            faucet_vip = self.vip_map(ipa)
+        return faucet_vip and ipa == faucet_vip.ip
+
+    def ip_in_vip_subnet(self, ipa, faucet_vip=None):
+        """Return faucet_vip if IP in same IP network as a VIP on this VLAN."""
+        if faucet_vip is None:
+            faucet_vip = self.vip_map(ipa)
+        if faucet_vip:
+            if ipa not in (
+                    faucet_vip.network.network_address,
+                    faucet_vip.network.broadcast_address):
+                return faucet_vip
         return None
 
     def from_connected_to_vip(self, src_ip, dst_ip):
