@@ -6,14 +6,11 @@ docker rmi faucet/test-base
 docker images
 
 if [ "${MATRIX_SHARD}" = "sanity" ] ; then
-  FAUCET_TESTS="FaucetSanityTest"
-  # TODO: move to docker.
-  cd tests
-  PYTHONPATH=.. ./test_coverage.sh || exit 1
+  FAUCET_TESTS="-u FaucetSanityTest"
+  ./tests/run_unit_tests.sh || exit 1
   codecov || true
-  cd ..
 else
-  ALLTESTFILES="tests/faucet_mininet_test_unit.py clib/clib_mininet_test_unit.py"
+  ALLTESTFILES="tests/integration/mininet_tests.py clib/clib_mininet_tests.py"
   ALLTESTS=`grep -E -o "^class (Faucet[a-zA-Z0-9]+Test)" ${ALLTESTFILES}|cut -f2 -d" "|sort`
   declare -A sharded
 
@@ -36,6 +33,5 @@ echo Shard $MATRIX_SHARD: $FAUCETTESTS
 sudo docker run --privileged --sysctl net.ipv6.conf.all.disable_ipv6=0 \
   -v $HOME/.cache/pip:/var/tmp/pip-cache \
   -e FAUCET_TESTS="${FAUCET_TESTS}" \
-  -e CODECOV_TOKEN="${CODECOV_TOKEN}" \
   -t ${FAUCET_TEST_IMG} || exit 1
 exit 0
